import re
import os
import io
import shutil
import tempfile
import urllib
import typing
from typing import Callable, Iterable, Optional, Generator, ContextManager, Union, Dict, Literal, Tuple, List
from types import GeneratorType
from contextlib import ExitStack, contextmanager
from abc import ABC, abstractmethod
from hashlib import sha256
import xml.etree.ElementTree as ET
import numpy as np
import pandas as pd
import pyterrier as pt

if typing.TYPE_CHECKING:
    from collections.abc import Buffer # type: ignore[attr-defined]
else:
    Buffer = Union[bytes, bytearray, memoryview]


def coerce_dataframe(obj):
    if isinstance(obj, pd.DataFrame):
        return obj
    if isinstance(obj, GeneratorType):
        #its a generator, lets assume it generates dataframes
        rtr=[]
        for x in obj:
            assert isinstance(x, pd.DataFrame)
            rtr.append(x)
        return pd.concat(rtr)

def autoopen(filename, mode='rb', **kwargs):
    """
    A drop-in for open() that applies automatic compression for .gz, .bz2 and .lz4 file extensions
    """

    if filename.endswith(".gz"):
        import gzip
        return gzip.open(filename, mode, **kwargs)
    elif filename.endswith(".bz2"):
        import bz2
        return bz2.open(filename, mode, **kwargs)
    elif filename.endswith(".lz4"):
        import lz4.frame
        return lz4.frame.open(filename, mode, **kwargs)
    return open(filename, mode, **kwargs)

def find_files(dir):
    """
    Returns all the files present in a directory and its subdirectories

    :param dir: The directory containing the files

    :return: A list of the paths to the files
    """
    files = []
    for (dirpath, dirnames, filenames) in os.walk(dir, followlinks=True):
        for name in filenames:
            files.append(os.path.join(dirpath, name))
    return sorted(files)


@contextmanager
def _finalized_open_base(path: str, mode: str, open_fn: Callable) -> Generator[io.BufferedIOBase, None, None]:
    assert mode in ('b', 't') # must supply either binary or text mode
    dirname = os.path.dirname(path)
    prefix, suffix = os.path.splitext(os.path.basename(path))
    prefix = f'.{prefix}.'
    suffix = f'.tmp{suffix}' # last part of the suffix needed for autoopen
    try:
        fd, path_tmp = tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=dirname)
        os.close(fd) # mkstemp returns a low-level file descriptor... Close it and re-open the file the normal way
        with open_fn(path_tmp, f'w{mode}') as fout:
            yield fout
        os.chmod(path_tmp, 0o666) # default file umask
    except:
        try:
            os.remove(path_tmp)
        except Exception:
            pass # edge case: removing temp file failed. Ignore and just raise orig error
        raise

    os.replace(path_tmp, path)


def finalized_open(path: str, mode: str) -> ContextManager[io.BufferedIOBase]:
    """
    Opens a file for writing, but reverts it if there was an error in the process.

    :param path: Path of file to open
    :param mode: Either t or b, for text or binary mode

    Example:
        Returns a contextmanager that provides a file object, so should be used in a "with" statement. E.g.::

            with pt.io.finalized_open("file.txt", "t") as f:
                f.write("some text")
            # file.txt exists with contents "some text"

        If there is an error when writing, the file is reverted::

            with pt.io.finalized_open("file.txt", "t") as f:
                f.write("some other text")
                raise Exception("an error")
            # file.txt remains unchanged (if existed, contents unchanged; if didn't exist, still doesn't)
    """
    return _finalized_open_base(path, mode, open)


def finalized_autoopen(path: str, mode: str) -> ContextManager[io.BufferedIOBase]:
    """
    Opens a file for writing with ``autoopen``, but reverts it if there was an error in the process.

    :param path: Path of file to open
    :param mode: Either t or b, for text or binary mode

    Example:
        Returns a contextmanager that provides a file object, so should be used in a "with" statement. E.g.::

            with pt.io.finalized_autoopen("file.gz", "t") as f:
                f.write("some text")
            # file.gz exists with contents "some text"

        If there is an error when writing, the file is reverted::

            with pt.io.finalized_autoopen("file.gz", "t") as f:
                f.write("some other text")
                raise Exception("an error")
            # file.gz remains unchanged (if existed, contents unchanged; if didn't exist, still doesn't)
    """
    return _finalized_open_base(path, mode, autoopen)

def ok_filename(fname : str) -> bool:
    """
    Checks to see if a filename is valid.
    """
    BAD_CHARS = ':"%/<>^|?' + os.sep
    for c in BAD_CHARS:
        if c in fname:
            return False
    return True

def touch(fname : str, mode=0o666, dir_fd=None, **kwargs):
    """
    Eqiuvalent to touch command on linux.
    Implementation from https://stackoverflow.com/a/1160227
    """
    flags = os.O_CREAT | os.O_APPEND
    with os.fdopen(os.open(fname, flags=flags, mode=mode, dir_fd=dir_fd)) as f:
        os.utime(f.fileno() if os.utime in os.supports_fd else fname,
            dir_fd=None if os.supports_fd else dir_fd, **kwargs)


def read_results(filename : str, format="trec", topics : Optional[pd.DataFrame] = None, dataset : Optional[pt.datasets.Dataset] =None, **kwargs) -> pd.DataFrame:
    """
    Reads a file into a results dataframe.

    :param filename: The filename of the file to be read. Compressed files are handled automatically. A URL is also supported for the "trec" format.
    :param format: The format of the results file: one of "trec", "letor". Default is "trec".
    :param topics: If provided, will merge the topics to merge into the results. This is helpful for providing query text. Cannot be used in conjunction with dataset argument.
    :param dataset: If provided, loads topics from the dataset (or dataset ID) and merges them into the results. This is helpful for providing query text. Cannot be used in conjunction with dataset topics.
    :param kwargs: Other arguments for the internal method

    :return: dataframe with usual qid, docno, score columns etc

    Examples::

        # a dataframe of results can be used directly in a pt.Experiment
        pt.Experiment(
            [ pt.io.read_results("/path/to/baselines-results.res.gz") ],
            topics,
            qrels,
            ["map"]
        )

        # make a transformer from a results dataframe, include the query text
        first_pass = pt.Transformer.from_df( pt.io.read_results("/path/to/results.gz", topics=topics) )
        # make a max_passage retriever based on a previously saved results
        max_passage = (first_pass 
            >> pt.text.get_text(dataset)
            >> pt.text.sliding()
            >> pt.text.scorer()
            >> pt.text.max_passage()
        )

    """
    if format not in SUPPORTED_RESULTS_FORMATS:
        raise ValueError("Format %s not known, supported types are %s" % (format, str(SUPPORTED_RESULTS_FORMATS.keys())))
    if SUPPORTED_RESULTS_FORMATS[format][0] is None:
        raise ValueError("Format %s does not support reading" % format)
    results = SUPPORTED_RESULTS_FORMATS[format][0](filename, **kwargs) # type: ignore
    if dataset is not None:
        assert topics is None, "Cannot provide both dataset and topics"
        if isinstance(dataset, str):
            dataset = pt.get_dataset(dataset)
        topics = dataset.get_topics()
    if topics is not None:
        results = pd.merge(results, topics, how='left', on='qid')
    return results

def _read_results_letor(filename, labels=False):

    def _parse_line(line):
            # $line =~ s/(#.*)$//;
            # my $comment = $1;
            # my @parts = split /\s+/, $line;
            # my $label = shift @parts;
            # my %hash = map {split /:/, $_} @parts;
            # return ($label, $comment, %hash);
        line, comment = line.split("#")
        line = line.strip()
        parts = re.split(r'\s+|:', line)
        label = parts.pop(0)
        m = re.search(r'docno\s?=\s?(\S+)', comment)
        docno = m.group(1)
        kv = {}
        qid = None
        for i, k in enumerate(parts):
            if i % 2 == 0:
                if k == "qid":
                    qid = parts[i+1]
                else:
                    kv[int(k)] = float(parts[i+1])
        features = np.array([kv[i] for i in sorted(kv.keys())])
        return (label, qid, docno, features)

    with autoopen(filename, 'rt') as f:
        rows = []
        for line in f:
            if line.startswith("#"):
                continue
            (label, qid, docno, features) = _parse_line(line)
            if labels:
                rows.append([qid, docno, features, label])
            else:
                rows.append([qid, docno, features])
        return pd.DataFrame(rows, columns=["qid", "docno", "features", "label"] if labels else ["qid", "docno", "features"])

def _read_results_trec(filename):
    df = pd.read_csv(filename, sep=r'\s+', names=["qid", "iter", "docno", "rank", "score", "name"], dtype={'qid': str, 'docno': str, 'rank': int, 'score': float}) 
    df = df.drop(columns="iter")
    return df

def write_results(res : pd.DataFrame, filename : str, format : Literal['trec', 'letor', 'minimal'] ="trec", append=False, **kwargs):
    """
    Write a results dataframe to a file.

    :param res: A results dataframe, with usual columns of qid, docno etc
    :param filename: The filename of the file to be written. Compressed files are handled automatically.
    :param format: The format of the results file: one of "trec", "letor", "minimal"
    :param append: Append to an existing file. Defaults to False.
    :param kwargs: Other arguments for the internal method

    Supported Formats:
        * "trec" -- output columns are `$qid Q0 $docno $rank $score $runname, space separated`
        * "letor" -- This follows the LETOR and MSLR datasets, in that output columns are `$label qid:$qid [$fid:$value]+ # docno=$docno`
        * "minimal": output columns are `$qid $docno $rank`, tab-separated. This is used for submissions to the MSMARCO leaderboard.

    """
    if format not in SUPPORTED_RESULTS_FORMATS:
        raise ValueError("Format %s not known, supported types are %s" % (format, str(SUPPORTED_RESULTS_FORMATS.keys())))
    if SUPPORTED_RESULTS_FORMATS[format][1] is None:
        raise ValueError("Format %s does not support writing" % format)
    # convert generators to results
    res = coerce_dataframe(res)
    return SUPPORTED_RESULTS_FORMATS[format][1](res, filename, append=append, **kwargs)  # type: ignore

def _write_results_trec(res, filename, run_name="pyterrier", append=False):
        res_copy = res.copy()[["qid", "docno", "rank", "score"]]
        res_copy.insert(1, "Q0", "Q0")
        res_copy.insert(5, "run_name", run_name)
        res_copy.to_csv(filename, sep=" ", mode='a' if append else 'w', header=False, index=False)

def _write_results_minimal(res, filename, run_name="pyterrier", append=False):
        res_copy = res.copy()[["qid", "docno", "rank"]]
        res_copy.to_csv(filename, sep="\t", mode='a' if append else 'w', header=False, index=False)

def _write_results_letor(res, filename, qrels=None, default_label=0, append=False):
    if qrels is not None:
        res = res.merge(qrels, on=['qid', 'docno'], how='left').fillna(default_label)
    mode='wa' if append else 'wt'
    with autoopen(filename, mode) as f:
        for row in res.itertuples():
            values = row.features
            label = row.label if qrels is not None else default_label
            feat_str = ' '.join( [ '%i:%f' % (i+1,values[i]) for i in range(len(values)) ] )
            f.write("%d qid:%s %s # docno=%s\n" % (label, row.qid, feat_str, row.docno))

def read_topics(filename : str, format :Literal['trec', 'trecxml', 'singleline'] = "trec", **kwargs) -> pd.DataFrame:
    """
    Reads a file containing topics.

    :param filename: The filename of the topics file. A URL is supported for the "trec" and "singleline" formats.
    :param format: One of "trec", "trecxml" or "singleline". Default is "trec"

    :return: pandas.Dataframe with columns=['qid','query'], where both columns have type str.

    Supported Formats:
        * "trec" -- an SGML-formatted TREC topics file. Delimited by TOP tags, each having NUM and TITLE tags; DESC and NARR tags are skipped by default. Control using whitelist and blacklist kwargs
        * "trecxml" -- a more modern XML formatted topics file. Delimited by topic tags, each having number tags. query, question and narrative tags are parsed by default. Control using tags kwarg.
        * "singleline" -- one query per line, preceeded by a space or colon. Tokenised by default, use tokenise=False kwargs to prevent tokenisation.
    """
    if format is None:
        format = "trec"
    if format not in SUPPORTED_TOPICS_FORMATS:
        raise ValueError("Format %s not known, supported types are %s" % (format, str(SUPPORTED_TOPICS_FORMATS.keys())))
    return SUPPORTED_TOPICS_FORMATS[format](filename, **kwargs)

def _read_topics_trec(file_path, doc_tag="top", id_tag="num", whitelist=["title"], blacklist=["desc","narr"], tokenise=False) -> pd.DataFrame:
    assert not tokenise, "Tokenisation is not supported for SGML-formatted topics; use pt.terrier.rewrite.tokenise() in your pipeline if needed" 
    def _open(filename):
        if filename.startswith('http://') or filename.startswith('https://'):
            return pt.io.download_stream(filename)
        return pt.io.autoopen(filename, 'rt')
    
    with _open(file_path) as f:
        data = f.read()

    # Ensure line breaks before each tag (helps regex clarity)
    data = re.sub(r'(?=<\w+>)', '\n', data)

    # List of tags we want to auto-close when a new one starts
    auto_close_tags = ["num", "title", "desc", "narr", "smry", "con", "fac", "def", "dom", "head"]

    # obtained by inspection of TREC-1 topics
    prefix_words = {
        "num": "Number:",
        "desc": "Description:",
        "narr": "Narrative:",
        "smry": "Summary:",
        "con": "Concept(s):",
        "fac": "Factor(s):",
        "def": "Definition(s):",
        "dom": "Domain:",
        "title": "Topic:",
    }

    # For each tag, if it doesn’t have a closing tag, and a new tag starts,
    # we insert an implicit </tag>
    for tag in auto_close_tags:
        # Matches: <tag> ... <other_tag>
        pattern = rf'(<{tag}>.*?)(?=<(?:{"|".join(auto_close_tags + ["top", "/top"])}|$))'
        data = re.sub(pattern, rf'\1</{tag}>', data, flags=re.S | re.I)

    from bs4 import BeautifulSoup
    soup = BeautifulSoup(data, "html.parser")

    topics = []
    for top in soup.find_all(doc_tag):
        
        num = top.find(id_tag).get_text(strip=True)
        # replace Number: prefix if present
        if id_tag in prefix_words and num.startswith(prefix_words[id_tag]):
            num = num.replace(prefix_words[id_tag], "", 1).strip()

        # e.g. TREC-1 has num as "051" in the topics, but "51" in the qrels
        if num.startswith("0"):
            num = num.replace("0", "", 1)

        query = ""
        for tag in whitelist:
            tag_content = top.find(tag)
            if not tag_content:
                continue
            text = tag_content.get_text(strip=True)
            text = text.replace("\n", " ").replace("\r", " ")
            if tag in prefix_words and text.startswith(prefix_words[tag]):
                text = text.replace(prefix_words[tag], "", 1).strip()
            query += text + " "

        topics.append({
            "qid": num.strip(),
            "query": query.strip(),
        })
    df = pd.DataFrame(topics)
    return df[["qid", "query"]]

@typing.no_type_check
def _read_topics_trecxml(filename : str, tags : List[str] = ["query", "question", "narrative"], tokenise=False) -> pd.DataFrame:
    """
    Parse a file containing topics in TREC-like XML format

    Args:
        filename(str): The path to the topics file

    Returns:
        pandas.Dataframe with columns=['qid','query']
    """
    assert not tokenise, "Tokenisation is not supported for XML-formatted topics; use pt.terrier.rewrite.tokenise() in your pipeline if needed" 
    _tags=set(tags)
    topics=[]
    tree = ET.parse(filename)
    root = tree.getroot()
    for child in root.iter('topic'):
        try:
            qid = child.attrib["number"]
        except KeyError:
            qid = child.find("number").text
        query = ""
        for tag in child:
            if tag.tag in _tags:
                query += " " + tag.text
        topics.append((str(qid), query.strip()))
    return pd.DataFrame(topics, columns=["qid", "query"])

def _read_topics_singleline(filepath, contains_qid=True, tokenise=False) -> pd.DataFrame:
    """
    Parse a file containing topics, one per line. Supports reading direct from URLs.
    Uses Terrier's parser if tokenise=True.

    Args:
        file_path(str): The path to the topics file
        tokenise(bool): whether the query should be tokenised, using Terrier's standard Tokeniser.
            If you are using matchop formatted topics, this should be set to False.

    Returns:
        pandas.Dataframe with columns=['qid','query']
    """
    assert not tokenise, "Tokenisation is not supported for XML-formatted topics; use pt.terrier.rewrite.tokenise() in your pipeline if needed" 
    qid_counter = 0
    
    def _open(filepath):
        if filepath.startswith('http://') or filepath.startswith('https://'):
            return pt.io.download_stream(filepath)
        return pt.io.autoopen(filepath, 'rt')
    
    with _open(filepath) as f:
        rows = []
        for line in f:
            line = line.strip()
            if line.startswith("#"):
                continue
            if len(line) == 0:
                continue
            if contains_qid:
                m = re.match(r'^([^:\s]+)[\s:]+(.*)$', line)
                if m:
                    qid = m.group(1)
                    query = m.group(2)
            else:
                qid_counter += 1
                qid = str(qid_counter)
            rows.append([qid, query])
        return pd.DataFrame(rows, columns=["qid", "query"]) 

def read_qrels(file_path : str) -> pd.DataFrame:
    """
    Reads a file containing qrels (relevance assessments)

    :param file_path: The path to the qrels file.  A URL is also supported.

    :return: pandas.Dataframe with columns=['qid','docno', 'label']
        with column types string, string, and int
    """
    df = pd.read_csv(file_path,
                     sep=r'\s+',
                     names=["qid", "iter", "docno", "label"],
                     dtype={"qid": "str", "docno": "str"})
    df = df.drop(columns="iter")
    return df

SUPPORTED_TOPICS_FORMATS = {
    "trec" : _read_topics_trec,
    "trecxml" : _read_topics_trecxml,
    "singleline": _read_topics_singleline
}

SUPPORTED_RESULTS_FORMATS : Dict[str, Tuple[Optional[Callable[..., pd.DataFrame]], Optional[Callable[..., None]]]] = {
    "trec" : (_read_results_trec, _write_results_trec),
    "letor" : (_read_results_letor, _write_results_letor),
    "minimal" : (None, _write_results_minimal)
}


def pyterrier_home() -> str:
    """
    Returns pyterrier's home directory. By default this is ~/.pyterrier, but it can also be set with the PYTERRIER_HOME
    env variable.
    """
    if "PYTERRIER_HOME" in os.environ:
        home = os.environ["PYTERRIER_HOME"]
    else:
        home = os.path.expanduser('~/.pyterrier')
    if not os.path.exists(home):
        os.makedirs(home)
    return home


@contextmanager
def finalized_directory(path: str) -> Generator[str, None, None]:
    """Creates a directory, but reverts it if there was an error in the process."""
    dirname = os.path.dirname(path)
    prefix, suffix = os.path.splitext(os.path.basename(path))
    prefix = f'.{prefix}.'
    suffix = f'.tmp{suffix}' # keep final suffix/extension
    try:
        path_tmp = tempfile.mkdtemp(suffix=suffix, prefix=prefix, dir=dirname)
        yield path_tmp
        os.chmod(path_tmp, 0o777) # default directory umask
    except:
        try:
            shutil.rmtree(path_tmp)
        except:
            raise
        raise

    os.replace(path_tmp, path)


def download(url: str, path: str, *, expected_sha256: Optional[str] = None, verbose: bool = True, headers={}) -> None:
    """Downloads a file from a URL to a local path."""
    with finalized_open(path, 'b') as fout, \
         download_stream(url, expected_sha256=expected_sha256, verbose=verbose, headers = headers) as fin:
        while chunk := fin.read1():
            fout.write(chunk)


@contextmanager
def download_stream(
    url: str,
    *,
    expected_sha256: Optional[str] = None,
    headers: Optional[Dict[str, str]] = None,
    verbose: bool = True
) -> Generator[io.BufferedIOBase, None, None]:
    """Downloads a file from a URL to a stream."""
    with ExitStack() as stack:
        request = urllib.request.Request(url, headers=headers or {})
        fin = stack.enter_context(urllib.request.urlopen(request))
        if fin.status != 200:
            raise OSError(f'Unhandled status code: {fin.status}')

        if verbose:
            total = int(fin.headers.get('Content-Length', 0)) or None
            fin = stack.enter_context(TqdmReader(fin, total=total, desc=url))

        if expected_sha256 is not None:
            fin = stack.enter_context(HashReader(fin, expected=expected_sha256))

        yield fin


@contextmanager
def open_or_download_stream(
    path_or_url: str,
    *,
    expected_sha256: Optional[str] = None,
    headers: Optional[Dict[str, str]] = None,
    verbose: bool = True
) -> Generator[io.BufferedIOBase, None, None]:
    """Opens a file or downloads a file from a URL to a stream."""
    if path_or_url.startswith('http://') or path_or_url.startswith('https://'):
        with download_stream(path_or_url, headers=headers, expected_sha256=expected_sha256, verbose=verbose) as fin:
            yield fin
    elif os.path.isfile(path_or_url):
        with ExitStack() as stack:
            fin = stack.enter_context(open(path_or_url, 'rb'))

            if verbose:
                total = os.path.getsize(path_or_url)
                fin = stack.enter_context(TqdmReader(fin, total=total, desc=path_or_url))

            if expected_sha256 is not None:
                fin = stack.enter_context(HashReader(fin, expected=expected_sha256))

            yield fin
    else:
        raise OSError(f'path or url {path_or_url!r} not found') # error can occur here if protocol entrypoints were not found - try pip install .


class _NosyReader(io.BufferedIOBase, ABC):
    def __init__(self, reader: io.BufferedIOBase):
        self.reader = reader
        self.seek = self.reader.seek # type: ignore[method-assign]
        self.tell = self.reader.tell # type: ignore[method-assign]
        self.seekable = self.reader.seekable # type: ignore[method-assign]
        self.readable = self.reader.readable # type: ignore[method-assign]
        self.writable = self.reader.writable # type: ignore[method-assign]
        self.flush = self.reader.flush # type: ignore[method-assign]
        self.isatty = self.reader.isatty # type: ignore[method-assign]

    @abstractmethod
    def on_data(self, data: bytes) -> None:
        pass

    def read1(self, size: Optional[int] = None) -> bytes:
        if size is None:
            size = io.DEFAULT_BUFFER_SIZE
        chunk = self.reader.read1(size)
        self.on_data(chunk)
        return chunk

    def read(self, size: Optional[int] = None) -> bytes:
        if size is None:
            size = io.DEFAULT_BUFFER_SIZE
        chunk = self.reader.read(size)
        self.on_data(chunk)
        return chunk

    def close(self) -> None:
        self.reader.close()


class _NosyWriter(io.BufferedIOBase, ABC):
    def __init__(self, writer: io.BufferedIOBase):
        self.writer = writer
        self.seek = self.writer.seek # type: ignore[method-assign]
        self.tell = self.writer.tell # type: ignore[method-assign]
        self.seekable = self.writer.seekable # type: ignore[method-assign]
        self.readable = self.writer.readable # type: ignore[method-assign]
        self.writable = self.writer.writable # type: ignore[method-assign]
        self.flush = self.writer.flush # type: ignore[method-assign]
        self.isatty = self.writer.isatty # type: ignore[method-assign]
        self.close = self.writer.close # type: ignore[method-assign]
        self.sha256 = sha256()

    @abstractmethod
    def on_data(self, data: Buffer) -> None:
        pass

    def write(self, data: Buffer) -> int:
        res = self.writer.write(data)
        self.on_data(data)
        return res

    def replace_writer(self, writer: io.BufferedIOBase) -> None:
        self.writer = writer # type: ignore[method-assign]
        self.seek = self.writer.seek # type: ignore[method-assign]
        self.tell = self.writer.tell # type: ignore[method-assign]
        self.seekable = self.writer.seekable # type: ignore[method-assign]
        self.readable = self.writer.readable # type: ignore[method-assign]
        self.writable = self.writer.writable # type: ignore[method-assign]
        self.flush = self.writer.flush # type: ignore[method-assign]
        self.isatty = self.writer.isatty # type: ignore[method-assign]
        self.close = self.writer.close # type: ignore[method-assign]


class HashReader(_NosyReader):
    """A reader that computes the sha256 hash of the data read."""
    def __init__(self, reader: io.BufferedIOBase, *, hashfn: Callable = sha256, expected: Optional[str] = None):
        """Create a HashReader."""
        super().__init__(reader)
        self.hash = hashfn()
        self.expected = expected

    def on_data(self, data: bytes) -> None:
        """Called when data is read."""
        self.hash.update(data)

    def hexdigest(self) -> str:
        """Return the hexdigest of the hash."""
        return self.hash.hexdigest()

    def close(self) -> None:
        """Close the reader and check the hash."""
        self.reader.close()
        if self.expected is not None:
            if self.expected.lower() != self.hexdigest():
                raise ValueError(f'Expected sha256 {self.expected!r} but found {self.hexdigest()!r}')


class HashWriter(_NosyWriter):
    """A writer that computes the sha256 hash of the data written."""
    def __init__(self, writer: io.BufferedIOBase, *, hashfn: Callable = sha256):
        """Create a HashWriter."""
        super().__init__(writer)
        self.hash = hashfn()

    def on_data(self, data: Buffer) -> None:
        """Called when data is written."""
        self.hash.update(data)

    def hexdigest(self) -> str:
        """Return the hexdigest of the hash."""
        return self.hash.hexdigest()


class TqdmReader(_NosyReader):
    """A reader that displays a progress bar."""
    def __init__(self, reader: io.BufferedIOBase, *, total: Optional[int] = None, desc: Optional[str] = None, disable: bool = False):
        """Create a TqdmReader."""
        super().__init__(reader)
        self.pbar = pt.tqdm(total=total, desc=desc, unit="B", unit_scale=True, unit_divisor=1024, disable=disable) # type: ignore[misc]

    def on_data(self, data: bytes) -> None:
        """Called when data is read."""
        self.pbar.update(len(data))

    def close(self) -> None:
        """Close the reader and the progress bar."""
        super().close()
        self.reader.close()
        self.pbar.close()


class CallbackReader(_NosyReader):
    """A reader that calls a callback with the data read."""
    def __init__(self, reader: io.BufferedIOBase, callback: Callable):
        """Create a CallbackReader."""
        super().__init__(reader)
        self.callback = callback

    def on_data(self, data: bytes) -> None:
        """Called when data is read."""
        self.callback(data)


class MultiReader(io.BufferedIOBase):
    """A reader that reads from multiple readers in sequence."""
    def __init__(self, readers: Iterable[io.BufferedIOBase]):
        """Create a MultiReader."""
        self.readers = iter(readers)
        self._reader: Optional[io.BufferedIOBase] = next(self.readers)
        self.reader: Optional[io.BufferedIOBase] = self._reader.__enter__()
        self.seek = self.reader.seek # type: ignore[method-assign]
        self.tell = self.reader.tell # type: ignore[method-assign]
        self.seekable = self.reader.seekable # type: ignore[method-assign]
        self.readable = self.reader.readable # type: ignore[method-assign]
        self.writable = self.reader.writable # type: ignore[method-assign]
        self.flush = self.reader.flush # type: ignore[method-assign]
        self.isatty = self.reader.isatty # type: ignore[method-assign]
        self.close = self.reader.close # type: ignore[method-assign]

    def read1(self, size: Optional[int] = None) -> bytes:
        """Read a single chunk of data."""
        if self.reader is None:
            raise RuntimeError('reader is closed')
        if size is None:
            size = io.DEFAULT_BUFFER_SIZE
        chunk = self.reader.read1(min(size, io.DEFAULT_BUFFER_SIZE))
        if len(chunk) == 0:
            self.reader.close()
            try:
                self._reader = next(self.readers)
            except StopIteration:
                self._reader = None
                self.reader = None
                return chunk
            self.reader = self._reader.__enter__()
            if hasattr(self.reader, 'pbar'):
                self.pbar = self.reader.pbar
            self.seek = self.reader.seek # type: ignore[method-assign]
            self.tell = self.reader.tell # type: ignore[method-assign]
            self.seekable = self.reader.seekable # type: ignore[method-assign]
            self.readable = self.reader.readable # type: ignore[method-assign]
            self.writable = self.reader.writable # type: ignore[method-assign]
            self.flush = self.reader.flush # type: ignore[method-assign]
            self.isatty = self.reader.isatty # type: ignore[method-assign]
            self.close = self.reader.close # type: ignore[method-assign]
            chunk = self.reader.read1(min(size, io.DEFAULT_BUFFER_SIZE))
        return chunk

    def read(self, size: Optional[int] = None) -> bytes:
        """Read data."""
        chunk = b''
        if size is None:
            size = io.DEFAULT_BUFFER_SIZE
        while len(chunk) < size and self.reader is not None:
            chunk += self.reader.read(size - len(chunk))
            if len(chunk) < size:
                self.reader.close()
                try:
                    self._reader = next(self.readers)
                except StopIteration:
                    self._reader = None
                    self.reader = None
                    return chunk
                self.reader = self._reader.__enter__()
                if hasattr(self.reader, 'pbar'):
                    self.pbar = self.reader.pbar
                self.seek = self.reader.seek # type: ignore[method-assign]
                self.tell = self.reader.tell # type: ignore[method-assign]
                self.seekable = self.reader.seekable # type: ignore[method-assign]
                self.readable = self.reader.readable # type: ignore[method-assign]
                self.writable = self.reader.writable # type: ignore[method-assign]
                self.flush = self.reader.flush # type: ignore[method-assign]
                self.isatty = self.reader.isatty # type: ignore[method-assign]
                self.close = self.reader.close # type: ignore[method-assign]
        return chunk


def path_is_under_base(path: str, base: str) -> bool:
    """Returns True if the path is under the base directory."""
    return os.path.realpath(os.path.abspath(os.path.join(base, path))).startswith(os.path.realpath(base))



from .base import BaseTestCase
import pandas as pd
import pyterrier as pt

class TestModel(BaseTestCase):

    def test_empty(self):
        df = pt.new.empty_Q()
        self.assertTrue("qid" in df.columns)
        self.assertTrue("query" in df.columns)

    def test_newR1(self):
        df = pt.new.ranked_documents([[1]])
        self.assertEqual(1, len(df))
        self.assertTrue("qid" in df.columns)
        self.assertEqual("1", df.iloc[0]["qid"])
        self.assertEqual("d1", df.iloc[0]["docno"])
        self.assertEqual(1, df.iloc[0]["score"])
        self.assertEqual(pt.model.FIRST_RANK, df.iloc[0]["rank"])

        df = pt.new.ranked_documents([[1]], docno=[["d100"]])
        self.assertEqual(1, len(df))
        self.assertTrue("qid" in df.columns)
        self.assertEqual("1", df.iloc[0]["qid"])
        self.assertEqual("d100", df.iloc[0]["docno"])
        self.assertEqual(1, df.iloc[0]["score"])
        self.assertEqual(pt.model.FIRST_RANK, df.iloc[0]["rank"])

        df = pt.new.ranked_documents([[2,1], [2,1], [2,1]])
        self.assertEqual(6, len(df))
        qids=["a", "b", "c"]
        df = pt.new.ranked_documents([[2,1], [2,1], [2,1]], qid=qids)
        self.assertEqual(6, len(df))
        offset=0
        for qid in qids:
            for i in [0,1]:
                self.assertEqual(df.iloc[offset]["qid"], qid)
                offset +=1
 
    def test_newR1_with_qid(self):
        df = pt.new.ranked_documents([[1]], qid=["q1"])
        self.assertEqual(1, len(df))
        self.assertTrue("qid" in df.columns)
        self.assertEqual("q1", df.iloc[0]["qid"])
        self.assertEqual("d1", df.iloc[0]["docno"])
        self.assertEqual(1, df.iloc[0]["score"])
        self.assertEqual(pt.model.FIRST_RANK, df.iloc[0]["rank"])

        df = pt.new.ranked_documents([[1]], docno=[["d100"]], qid=["q1"])
        self.assertEqual(1, len(df))
        self.assertTrue("qid" in df.columns)
        self.assertEqual("q1", df.iloc[0]["qid"])
        self.assertEqual("d100", df.iloc[0]["docno"])
        self.assertEqual(1, df.iloc[0]["score"])
        self.assertEqual(pt.model.FIRST_RANK, df.iloc[0]["rank"])

    def test_newR1_with_docid(self):
        df = pt.new.ranked_documents([[1]], qid=["q1"], docid=[[0]])
        self.assertEqual(1, len(df))
        self.assertTrue("qid" in df.columns)
        self.assertEqual("q1", df.iloc[0]["qid"])
        self.assertEqual("d1", df.iloc[0]["docno"])
        self.assertEqual(0, df.iloc[0]["docid"])
        self.assertEqual(1, df.iloc[0]["score"])
        self.assertEqual(pt.model.FIRST_RANK, df.iloc[0]["rank"])

    def test_new_1query_no_qid(self):
        df = pt.new.queries("a")
        self.assertEqual(1, len(df))
        self.assertTrue("qid" in df.columns)
        self.assertTrue("query" in df.columns)
        self.assertEqual("1", df.iloc[0]["qid"])
        self.assertEqual("a", df.iloc[0]["query"])
        
    def test_new_1query_w_qid(self):
        df = pt.new.queries("a", "2")
        self.assertEqual(1, len(df))
        self.assertTrue("qid" in df.columns)
        self.assertTrue("query" in df.columns)
        self.assertEqual("2", df.iloc[0]["qid"])
        self.assertEqual("a", df.iloc[0]["query"])

    def test_new_2query_w_qid(self):
        df = pt.new.queries(["a", "b"], ["2", "1"])
        self.assertEqual(2, len(df))
        self.assertTrue("qid" in df.columns)
        self.assertTrue("query" in df.columns)
        self.assertEqual("2", df.iloc[0]["qid"])
        self.assertEqual("a", df.iloc[0]["query"])
        self.assertEqual("1", df.iloc[1]["qid"])
        self.assertEqual("b", df.iloc[1]["query"])

    def test_new_2query_w_qid_hist(self):
        df = pt.new.queries(["a", "b"], ["2", "1"], hist=["cc", "dd"])
        self.assertEqual(2, len(df))
        self.assertTrue("qid" in df.columns)
        self.assertTrue("query" in df.columns)
        self.assertTrue("hist" in df.columns)
        self.assertEqual("2", df.iloc[0]["qid"])
        self.assertEqual("a", df.iloc[0]["query"])
        self.assertEqual("1", df.iloc[1]["qid"])
        self.assertEqual("b", df.iloc[1]["query"])
        self.assertEqual("cc", df.iloc[0]["hist"])
        self.assertEqual("dd", df.iloc[1]["hist"])
        


"""This module provides useful utility methods for getting information about PyTerrier objects.

.. note::
    This is an advanced module that is not typically used by end users.
"""
import enum
import inspect
import dataclasses
from typing import Any, Dict, List, Literal, Optional, Protocol, Type, Tuple, Union, cast, overload, runtime_checkable

import pandas as pd

import pyterrier as pt

__all__ = [
    'artifact_type_format',
    'transformer_inputs',
    'transformer_outputs',
    'transformer_attributes',
    'transformer_apply_attributes',
    'subtransformers',
    'InspectError',
    'TransformerAttribute',
    'HasTransformInputs',
    'HasTransformOutputs',
    'HasAttributes',
    'HasApplyAttributes',
    'HasSubtransformers',
]


class InspectError(TypeError):
    """Base exception for inspection errors."""
    pass


def artifact_type_format(
    artifact: Union[Type, 'pt.Artifact'],
    *,
    strict: bool = True,
) -> Optional[Tuple[str, str]]:
    """Returns the type and format of the specified artifact.

    These values are sourced by either the ``ARTIFACT_TYPE`` and ``ARTIFACT_FORMAT`` constants of the artifact, or (if these
    are not available) by matching on the entry points.

    Args:
        artifact: The artifact to inspect.
        strict: If True, raises an error if the artifact's type or format could not be determined.

    Returns:
        A tuple containing the artifact's type and format, or ``None`` if the type and format could not be determined
        and ``strict==False``.

    Raises:
        InspectError: If the artifact's type or format could not be determined and ``strict==True``
    """
    artifact_type, artifact_format = None, None

    # Source #1: ARTIFACT_TYPE and ARTIFACT_FORMAT constants
    if hasattr(artifact, 'ARTIFACT_TYPE') and hasattr(artifact, 'ARTIFACT_FORMAT'):
        artifact_type = artifact.ARTIFACT_TYPE
        artifact_format = artifact.ARTIFACT_FORMAT

    # Source #2: entry point name
    if artifact_type is None or artifact_format is None:
        for entry_point in pt.utils.entry_points('pyterrier.artifact'):
            if artifact.__module__.split('.')[0] != entry_point.value.split(':')[0].split('.')[0]:
                continue # only try loading entry points that share the same top-level module
            entry_point_cls = entry_point.load()
            if isinstance(artifact, type) and artifact == entry_point_cls or isinstance(artifact, entry_point_cls):
                artifact_type, artifact_format = entry_point.name.split('.', 1)
                break

    if artifact_type is None or artifact_format is None:
        if strict:
            raise InspectError(f'{artifact} does not provide type and format (either as constants or via entry point)')
        return None

    return artifact_type, artifact_format

def indexer_inputs(
    indexer : pt.Indexer,
    *,
    strict : bool = True
) -> Optional[List[List[str]]]:
    """
    Infers supported input column configurations (a ``List[List[str]]``) for a pt.Indexer instance.
    Orthogonal to ``transformer_inputs``. This implementation inspects the ``index_inputs()``
    method of the indexer, as other methods of transformer inspection arent applicable to indexers.

    Args:
        indexer: An instance of the indexer to inspect.
        strict: If True, raises an error if the indexer cannot be inferred. If False, returns
            None in these cases.

    Returns:
        A list of input column configurations (``List[List[str]]``) accepted by this indexer.

    Raises:
        InspectError: If the indexer cannot be inspected and ``strict==True``.
    """
    result = indexer.index_inputs()
    if result is None or not isinstance(result[0], list) or (len(result[0]) > 0 and not isinstance(result[0][0], str)):
        if strict:
            msg = f"Cannot determine inputs for {indexer} - index_inputs() should be implemented"
            raise InspectError(msg)
        return None
    return result

def transformer_inputs(
    transformer: pt.Transformer,
    *,
    strict: bool = True,
) -> Optional[List[List[str]]]:
    """Infers supported input column configurations (a ``List[List[str]]``) for a transformer.

    The method tries to infer the input columns that the transformer accepts by calling it with an empty DataFrame and inspecting
    a resulting ``pt.validate.InputValidationError``. If the transformer does not raise an error, it tries to infer the input columns
    by calling it with a pre-defined set of input columns.

    To handle edge cases, you can implement the :class:`~pyterrier.inspect.HasTransformInputs` protocol, which allows you to define a custom
    ``transform_inputs`` method that returns a list of input column configurations accepted by the transformer. ``transform_inputs``
    can also be an attribute instead of a method. In this case, it can be a list of lists of input columns (i.e., a list of valid
    input column configurations). Note that ``transform_inputs`` is allowed to return a ``List[str]``. If this is the case, it is converted
    to a ``List[List[str]]`` automatically.

    The list of input specifications is assumed to be prioritized. For instance, schematics will show the first valid specification
    when multiple are valid for the pipeline.

    Args:
        transformer: An instance of the transformer to inspect.
        strict: If True, raises an error if the transformer cannot be inferred or are not accepted. If False, returns
            None in these cases.

    Returns:
        A list of input column configurations (``List[List[str]]``) accepted by this transformer.

    Raises:
        InspectError: If the transformer cannot be inspected and ``strict==True``.
    """
    result : List[List[str]] = []
    received = None
    if isinstance(transformer, HasTransformInputs) and transformer.transform_inputs is not None:
        ext_result : Union[List[str], List[List[str]]]
        if not callable(transformer.transform_inputs):
            ext_result = transformer.transform_inputs
            received = "transformer.transform_inputs attribute"
        else:
            try:
                ext_result = transformer.transform_inputs()
                received = "transformer.transform_inputs() method"
            except Exception as ex:
                if strict:
                    raise InspectError(f"Cannot determine inputs for {transformer}") from ex
                else:
                    return None
        if len(ext_result) > 0:
            if isinstance(ext_result[0], str):
                ext_result = cast(List[str], ext_result) # noqa: PT100 (this is typing.cast, not jinus.cast)
                result = [ext_result] # convert to a List[List[str]]
            else:
                ext_result = cast(List[List[str]], ext_result) # noqa: PT100 (this is typing.cast, not jinus.cast)
                result = ext_result
    else:
        try:
            transformer(pd.DataFrame())
            # if this succeeds without an error, the transformer accepts frames without any columns
            result = [[]]
        except pt.validate.InputValidationError as ive:
            result = [mode.missing_columns for mode in ive.modes]
            received = "validation using invocation on empty 0-cols frame"
        except Exception:
            for mode, frame_type in [
                (['qid', 'query'] , "Q"),
                (['qid', 'query', 'docno', 'score', 'rank'], "R"),
            ]:
                try:
                    transformer(pd.DataFrame(columns=mode))
                    result.append(mode)
                    received = f"validation using invocation on empty {frame_type} frame"
                except Exception:
                    continue
    if not isinstance(result, list) or len(result) == 0:
        if strict:
            msg = f"Cannot determine inputs for {transformer}"
            if received is not None:
                msg += f"received by {received}: {result}"
            else:
                msg += " - no inspections succeeded"
            raise InspectError(msg)
        return None
    if not isinstance(result[0], list) or (len(result[0]) > 0 and not isinstance(result[0][0], str)):
        if strict:
            raise InspectError(f"Cannot determine inputs for {transformer} - invalid columns specified by {received}: {result}")
        return None
    return result

@overload
def transformer_outputs(
    transformer: pt.Transformer,
    input_columns: List[str],
    *,
    strict: Literal[True] = ...,
) -> List[str]: ...

@overload
def transformer_outputs(
    transformer: pt.Transformer,
    input_columns: List[str],
    *,
    strict: Literal[False] = ...,
) -> Optional[List[str]]: ...

def transformer_outputs(
    transformer: pt.Transformer,
    input_columns: List[str],
    *,
    strict: bool = True,
) -> Optional[List[str]]:
    """Infers the output columns for a transformer based on the provided input columns.

    If the transformer implements the :class:`~pyterrier.inspect.HasTransformOutputs` protocol,
    the method calls its ``transform_outputs`` method to determine the output columns. If the transformer does not
    implement the protocol, it attempts to infer the output columns by calling the transformer with an empty DataFrame.

    Args:
        transformer: An instance of the transformer to inspect.
        input_columns: A list of the columns present in the input frame.
        strict: If True, raises an error if the transformer cannot be inferred or are not accepted. If False, returns
            None in these cases.

    Returns:
        A list of the columns present in the output for ``transformer`` given ``input_columns``.

    Raises:
        InspectError: If the transformer's outputs could not be determined and ``strict==True``.
        pt.validate.InputValidationError: If input validation fails in the transformer and ``strict==True``.
    """
    if isinstance(transformer, HasTransformOutputs):
        try:
            return transformer.transform_outputs(input_columns)
        except pt.validate.InputValidationError:
            if strict:
                raise
            else:
                return None
        except Exception as ex:
            if strict:
                raise InspectError(f"Cannot determine outputs for {transformer} with inputs: {input_columns}") from ex
            else:
                return None

    try:
        res = transformer.transform(pd.DataFrame(columns=input_columns))
        return list(res.columns)
    except pt.validate.InputValidationError as ive:
        if strict:
            # add the underlying class to the IVE error, so its more clear whats not been validated
            # this improves readability for subtransformers 
            ive.args = (ive.args[0] + f" {transformer}", )
            raise ive
        else:
            return None
    except Exception as ex:
        if strict:
            raise InspectError(f"Cannot determine outputs for {transformer} with inputs: {input_columns}") from ex
        else:
            return None


@dataclasses.dataclass
class TransformerAttribute:
    """A dataclass representing an attribute of a transformer.

    Attributes:
        name: The name of the attribute.
        value: The value of the attribute.
        init_default_value: The default value of the attribute for the ``__init__`` method (if available) or ``inspect.Parameter.empty`` if not available.
        init_parameter_kind: The kind of the parameter in the ``__init__`` method (if available) or ``None`` if not available.
    """
    def __init__(
        self,
        name: str,
        value: Any,
        init_default_value: Any = inspect.Parameter.empty,
        init_parameter_kind: Optional[inspect._ParameterKind] = None,
    ):
        # we need to define __init__ directly to avoid issues with sphinx thinking that self.init_parameter_kind is an alias to inspect.Parameter.empty
        self.name = name
        self.value = value
        self.init_default_value = init_default_value
        self.init_parameter_kind = init_parameter_kind

    name: str
    value: Any
    init_default_value: Any
    init_parameter_kind: Optional[inspect._ParameterKind]

    MISSING = object()


def transformer_attributes(transformer: pt.Transformer, *, strict: bool = True) -> List[TransformerAttribute]:
    """Infers a list of attributes of the transformer.

    Here, an attribute is defined as any attribute of the transformer that is explicitly set by the ``__init__`` method,
    either under the same name (e.g., ``self.foo = foo``) or as a private attribute (e.g., ``self._foo = foo``).

    This definition allows for a set of attributes that should describe the state of a transformer. These attributes can
    be used to reconstruct the transformer from its attributes, e.g., by calling :meth:`~pyterrier.inspect.transformer_apply_attributes`.

    To handle edge cases (e.g., where the ``__init__`` parameters do not match the attribute names), you can implement
    the :class:`~pyterrier.inspect.HasAttributes` protocol.

    Args:
        transformer: The transformer to inspect.
        strict: If True, raises an error if an attribute cannot be identified from the transformer. If False, the attribute's value is set to ``TransformerAttribute.MISSING`` in these cases.

    Returns:
        A list of :class:`~pyterrier.inspect.TransformerAttribute` objects representing the attributes of the transformer.

    Raises:
        InspectError: If the attributes cannot be identified from the transformer.
    """
    if isinstance(transformer, HasAttributes):
        return transformer.attributes()
    result = []
    signature = inspect.signature(transformer.__class__.__init__)
    for p in list(signature.parameters.values())[1:]: # [1:] to skip first arg ("self") which is bound to the instance.
        if p.name.startswith('_'):
            continue # Skip private constructor parameters
        if hasattr(transformer, f'_{p.name}'):
            val = getattr(transformer, f'_{p.name}')
        elif hasattr(transformer, p.name):
            val = getattr(transformer, p.name)
        else:
            if strict:
                raise InspectError(f"Cannot identify attribute {p.name} in transformer {transformer}. Ensure that the attribute is set in the __init__ method.")
            else:
                val = TransformerAttribute.MISSING # could not identify the attribute
        result.append(TransformerAttribute(
            name=p.name,
            value=val,
            init_default_value=p.default,
            init_parameter_kind=p.kind,
        ))
    return result


def transformer_apply_attributes(transformer: pt.Transformer, **kwargs: Any) -> pt.Transformer:
    """Returns a new transformer instance from the provided transformer and updated attributes (as keyword arguments).

    This method is useful for constructing new transformer with some attributes replaced. For instance, when implemeting
    methods like :meth:`~pyterrier.transformers.SupportsFuseRankCutoff.fuse_rank_cutoff`, you frequently need to replace the
    ``num_results`` attribute of a transformer with a new value while keeping the remainder of the attributes the same.

    This method uses :meth:`~pyterrier.inspect.transformer_attributes` to identify the attributes of the transformer and
    then applies the provided keyword arguments to the transformer attributes. The method then reconstructs the transformer
    by calling its ``__init__`` method with the updated attributes.

    To handle edge cases (e.g., where the ``__init__`` parameters do not match the attribute names), you can implement
    the :class:`~pyterrier.inspect.HasApplyAttributes` protocol.

    Args:
        transformer: The transformer to apply the attributes to.
        **kwargs: Keyword arguments representing the attributes to set on the transformer.

    Returns:
        A new instance of the transformer with the provided attributes applied.

    Raises:
        InspectError: If an attribute is not found in the transformer or if attributes cannot be identified from the transformer.
    """
    if isinstance(transformer, HasApplyAttributes):
        return transformer.apply_attributes(**kwargs)
    attributes = transformer_attributes(transformer)
    for attr in attributes:
        if attr.name in kwargs:
            attr.value = kwargs.pop(attr.name)
    if any(kwargs):
        raise InspectError(f"Unknown attributes {list(kwargs.keys())} for transformer {transformer}")
    missing_attributes = [attr.name for attr in attributes if attr.value is TransformerAttribute.MISSING]
    if missing_attributes:
        raise InspectError(f"Attributes {missing_attributes} for transformer {transformer} are missing.")
    init_args = []
    init_kwargs = {}
    for attr in attributes:
        if attr.init_parameter_kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            init_args.append(attr.value)
        elif attr.init_parameter_kind == inspect.Parameter.KEYWORD_ONLY:
            init_kwargs[attr.name] = attr.value
        elif attr.init_parameter_kind == inspect.Parameter.POSITIONAL_ONLY:
            init_args.append(attr.value)
        elif attr.init_parameter_kind == inspect.Parameter.VAR_POSITIONAL:
            init_args.extend(attr.value)
        elif attr.init_parameter_kind == inspect.Parameter.VAR_KEYWORD:
            init_kwargs.update(attr.value)
    return transformer.__class__(*init_args, **init_kwargs)


def subtransformers(transformer: pt.Transformer) -> Dict[str, Union[pt.Transformer, List[pt.Transformer]]]:
    """Infers a dictionary of subtransformers for the given transformer.

    A subtransformer is a transformer that is used by another transformer to complete its task. Examples
    include those used by caches (e.g., ``scorer`` in :class:`pyterrier_caching.ScorerCache`) and the list
    of transformers that are used by a :class:`pyterrier_alpha.fusion.RRFusion` transformer.

    If the transformer implements the :class:`~pyterrier.inspect.HasSubtransformers` protocol,
    the method calls its ``subtransformers`` method to retrieve the subtransformers. If the transformer does not
    implement the protocol, the method inspects the transformer to identify any attributes of a transformer that
    are instance of pt.Transformer (or list/tuple of Transformer), returning a dictionary where the keys
    where the keys are the names of the subtransformers and the values are the subtransformers themselves. If the
    transformer does not have any subtransformers, an empty dictionary is returned.

    Args:
        transformer: The transformer to inspect.

    Returns:
        A dictionary of the provided transformer's subtransformers.

    Raises:
        InspectError: If the subtransformers cannot be identified from the transformer.
    """
    if isinstance(transformer, HasSubtransformers):
        return transformer.subtransformers()
    result: Dict[str, Union[pt.Transformer, List[pt.Transformer]]] = {}
    for attr in transformer_attributes(transformer, strict=False):
        if isinstance(attr.value, pt.Transformer) and not isinstance(attr.value, pt.Artifact):
            result[attr.name] = attr.value
        elif isinstance(attr.value, (list, tuple)) and all(isinstance(v, pt.Transformer) and not isinstance(v, pt.Artifact) for v in attr.value):
            result[attr.name] = list(attr.value)
    return result

def _minimal_inputs(all_configs : List[Optional[List[List[str]]]]) -> Optional[List[List[str]]]:
    """
    Among all input configurations, find the minimal common subset(s) of attributes needed
    for success invocation.
    """
    # if any configs are unknown, the minimal inputs is unknown
    if any([config is None for config in all_configs]):
        return None
    
    # essentially a cast:
    non_optional: List[List[List[str]]] = [config for config in all_configs if config is not None]
    all_configs_sets = [ 
        set(a) for tconfig in non_optional for a in tconfig
    ]
    
    from itertools import chain, combinations
    # universe of all columns
    universe = set(chain.from_iterable(all_configs_sets))

    # test subsets in increasing size
    plausible = []
    for r in range(1, len(universe) + 1):
        for subset in combinations(universe, r):
            ssubset = set(subset)
            # Check if subset works for all objects
            if all(any(set(schema).issubset(ssubset) for schema in obj) for obj in non_optional):
                plausible.append(list(ssubset))
    return plausible

class TransformerType(enum.Flag):
    """An enum representing the type of a transformer."""
    transformer = enum.auto()
    indexer = enum.auto()


def transformer_type(transformer: pt.Transformer) -> TransformerType:
    """Returns the type of the transformer as a :class:`~pyterrier.inspect.TransformerType` flag enum.

    The type can be one of:
    - ``TransformerType.transformer``: The transformer is a :class:`~pyterrier.Transformer` but not an :class:`~pyterrier.Indexer`.
    - ``TransformerType.indexer``: The transformer is an :class:`~pyterrier.Indexer` but does not implement ``transform`` or ``transform_iter``.
    - ``TransformerType.transformer | TransformerType.indexer``: The transformer is both a :class:`~pyterrier.Transformer` and an :class:`~pyterrier.Indexer`.
    - ``TransformerType(0)``: The transformer is neither a :class:`~pyterrier.Transformer` nor an :class:`~pyterrier.Indexer`.

    Args:
        transformer: The transformer to inspect.

    Returns:
        A :class:`~pyterrier.inspect.TransformerType` flag representing the type of the transformer.
    """
    if isinstance(transformer, pt.Indexer):
        if transformer.__class__.transform != pt.Indexer.transform or transformer.__class__.transform_iter != pt.Indexer.transform_iter:
            # Indexer that also implements transform or transform_iter (or both)
            return TransformerType.transformer | TransformerType.indexer # both a Transformer and an Indexer
        else:
            # Indexer that doesn't implement transform or transform_iter
            return TransformerType.indexer # only Indexer
    if isinstance(transformer, pt.Transformer):
        return TransformerType.transformer # only Transformer
    return TransformerType(0) # neither Transformer nor Indexer


@runtime_checkable
class HasTransformInputs(Protocol):
    """Protocol for transformers that provide a ``transform_inputs`` method.

    ``transform_inputs`` allows for inspection of the inputs accepted by transformers without needing to run it.

    When this method is present in a :class:`~pyterrier.Transformer` object, it must return either:
    
    - A list of lists of input columns (i.e., a list of valid input column configurations)
    - A list of input columns (i.e., a single valid input column configuration)

    If the input columns of the transformer do not depend on the instance, ``transform_inputs`` can also be
    an attribute with a value of type ``List[str]`` or ``List[List[str]]``.

    If ``transform_inputs is None``, it is ignored.

    This method need not be present in a Transformer class - it is an optional extension;
    an alternative is that the input columns are determined by calling the transformer with an empty ``DataFrame``.

    .. code-block:: python
        :caption: Example ``transform_inputs`` function, implementing :class:`~pyterrier.inspect.HasTransformInputs`.

        class MyRetriever(pt.Transformer):

            def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
                pt.validate.query_frame(inp, ['query'])
                # ... perform retrieval ...
                # return the same columns as inp plus docno, score, and rank. E.g., using DataFrameBuilder.

            def transform_inputs(self) -> Union[List[str], List[List[str]]]:
                # NOTE: This method isn't required in this case, since inspect will be able to infer required
                # columns from pt.validate. It's just a demonstration.
                return ['qid', 'query']

    """
    def transform_inputs(self) -> Union[List[List[str]], List[str]]:
        """Returns a list of input columns accepted by the transformer.

        Returns:
            Input column configuration(s) accepted by this transformer.
        """



@runtime_checkable
class HasTransformOutputs(Protocol):
    """Protocol for transformers that provide a ``transform_outputs`` method.

    ``transform_outputs`` allows for inspection of the outputs of transformers without needing to run it.

    When this method is present in a :class:`~pyterrier.Transformer` object, it must return a
    list of the output columns present given the provided input columns or raise an ``InputValidationError``
    if the inputs are not accepted by the transformer.

    This method need not be present in a Transformer class - it is an optional extension;
    an alternative is that the output columns are determined by calling the transformer
    with an empty ``DataFrame``.

    Due to risks and maintanence burden in ensuring that ``transform`` and ``transform_outputs`` behave identically,
    it is recommended to only implement ``transform_outputs`` when calling the transformer with an empty DataFrame to
    inspect the behavior is undesireable, e.g., if calling the transformer is expensive.

    .. code-block:: python
        :caption: Example ``transform_outputs`` function, implementing :class:`~pyterrier.inspect.HasTransformOutputs`.

        class MyRetriever(pt.Transformer):

            def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
                pt.validate.query_frame(inp, ['query'])
                # ... perform retrieval ...
                # return the same columns as inp plus docno, score, and rank. E.g., using DataFrameBuilder.

            def transform_outputs(self, input_columns: List[str]) -> List[str]:
                pt.validate.query_frame(input_columns, ['query'])
                return input_columns + ['docno', 'score', 'rank']

    """
    def transform_outputs(self, input_columns: List[str]) -> List[str]:
        """Returns a list of the output columns present given the ``input_columns``.

        The method must return exactly the same output columns as ``transform`` would given the provided input
        columns. If the input columns are not accepted by the transformer, the method should raise an
        ``InputValidationError`` (e.g., through ``pt.validate``).

        Args:
            input_columns: A list of the columns present in the input frame.

        Returns:
            A list of the columns present in the output for this transformer given ``input_columns``.

        Raises:
            pt.validate.InputValidationError: If the input columns are not accepted by the transformer.
            pt.inspect.InspectError: If the transformer is uninspectable.
        """


@runtime_checkable
class HasAttributes(Protocol):
    """Protocol for transformers that provide an ``attributes`` method.

    ``attributes`` allows for identifying the attributes of a transformer without needing to traverse
    its attributes manually.

    When this method is present in a :class:`~pyterrier.Transformer` object, it must return a
    list of :class:`~pyterrier.inspect.TransformerAttribute` objects, where each object represents an attribute of the
    transformer and corresponding metadata about how the attribute is assigned.

    This method need not be present in a Transformer class - it is an optional extension.
    """
    def attributes(self) -> List[TransformerAttribute]:
        """Returns a list of attributes of the transformer."""


@runtime_checkable
class HasApplyAttributes(Protocol):
    """Protocol for transformers that provide an ``apply_attributes`` method.

    ``apply_attributes`` returns a new transformer with updated attributes (as keyword arguments).

    This method need not be present in a Transformer class - it is an optional extension.
    """
    def apply_attributes(self, **kwargs: Any) -> pt.Transformer:
        """Returns a new transformer instance from the provided transformer and updated attributes (as keyword arguments)."""



@runtime_checkable
class HasSubtransformers(Protocol):
    """Protocol for transformers that provide a ``subtransformers`` method.

    ``subtransformers`` allows for identifying subtransformers of a transformer without needing to traverse
    its attributes manually.

    When this method is present in a :class:`~pyterrier.Transformer` object, it must return a
    dict where the keys are the names of the subtransformers and the values are the subtransformers (or list
    of subtransformers) themselves.

    This method need not be present in a Transformer class - it is an optional extension. See :meth:`pyterrier.inspect.subtransformers`
    for the default implementation.
    """
    def subtransformers(self) -> Dict[str, Union[pt.Transformer, List[pt.Transformer]]]:
        """Returns a dictionary of subtransformers for the transformer.

        The method must return a dictionary where the keys are the names of the subtransformers and the values are
        the subtransformers themselves. If the transformer does not have any subtransformers, an empty dictionary
        should be returned.
        """



import unittest
import pyterrier as pt
from .base import BaseTestCase

class TestDatasets(BaseTestCase):

    def test_list_datasets(self):
        df = pt.list_datasets()
        self.assertIsNotNone(df)
        self.assertTrue(len(df) > 2)

    def test_webtrack_gov(self):
        import pyterrier as pt
        import requests, urllib
        try:
            for k in ["trec-wt-2002", "trec-wt-2003", "trec-wt-2004"]:
                ds = pt.get_dataset(k)
                for t in ["np", "td", "hp"]:
                    if k != "trec-wt-2004":
                        #HP finding only for the 2004 task?
                        continue
                    topics = ds.get_topics(t)
                    qrels = ds.get_qrels(t)
                    
                    #check that the qrels qid match the topics.
                    join = topics.merge(qrels, on=["qid"])
                    self.assertTrue(len(join) > 0)
        except requests.exceptions.ConnectionError:
            self.skipTest("NIST not reachable")
        except urllib.error.URLError:
            self.skipTest("NIST not reachable")

    def test_webtrack_cw09(self):
        import pyterrier as pt
        import requests, urllib
        try:
            for k in ["trec-wt-2009", "trec-wt-2010", "trec-wt-2011", "trec-wt-2012"]:
                ds = pt.get_dataset(k)
                topics = ds.get_topics()
                qrels = ds.get_qrels("adhoc")
                
                # check that the qrels match the topics.
                join = topics.merge(qrels, on=["qid"])
                self.assertTrue(len(join) > 0)
        except requests.exceptions.ConnectionError:
            self.skipTest("NIST not reachable")
        except urllib.error.URLError:
            self.skipTest("NIST not reachable")
        except requests.exceptions.HTTPError:
            self.skipTest("NIST not reachable")
    
    def test_vaswani_corpus_iter(self):
        import pyterrier as pt
        dataset = pt.datasets.get_dataset("vaswani")
        self.assertIsNotNone(dataset)
        iter = dataset.get_corpus_iter()
        self.assertIsNotNone(iter)
        doc = next(iter)
        self.assertEqual(doc["docno"], "1")
        self.assertTrue(doc["text"].startswith("compact memories have flexible capacities"))

    def test_vaswani_from_dataset(self):
        import pyterrier as pt
        dataset = pt.datasets.get_dataset("vaswani")
        br = pt.terrier.Retriever.from_dataset(dataset)
        br.search("chemical reactions")
        
    def test_vaswani(self):
        import pyterrier as pt
        dataset = pt.datasets.get_dataset("vaswani")
        self.assertIsNotNone(dataset)

        topics = dataset.get_topics()
        self.assertIsNotNone(topics)
        self.assertEqual(len(topics), 93)

        qrels = dataset.get_qrels()
        self.assertIsNotNone(qrels)
        self.assertEqual(len(qrels), 2083)

        indexref = dataset.get_index()
        self.assertIsNotNone(indexref)
        with pt.IndexFactory.of(indexref) as index:
            self.assertIsNotNone(index)
            self.assertEqual(index.getCollectionStatistics().getNumberOfDocuments(), 11429)

        # do it once again, to ensure it works locally
        dataset = pt.datasets.get_dataset("vaswani")
        topics = dataset.get_topics()
        self.assertIsNotNone(topics)
        self.assertEqual(len(topics), 93)

        # test the newer get_topicsqrels
        pt.Experiment([pt.terrier.Retriever(dataset.get_index())], *dataset.get_topicsqrels(), ["map"])

if __name__ == "__main__":
    unittest.main()



from .base import BaseTestCase
import pandas as pd
from pyterrier.model import add_ranks, FIRST_RANK, coerce_queries_dataframe, coerce_dataframe_types, split_df
import pyterrier as pt
class TestModel(BaseTestCase):

    def test_R_to_Q(self):
        df1 = pt.new.ranked_documents([[1, 2], [2,0]], query=[["a a", "a a"], ["b b", "b b"]])
        df2 = pt.model.ranked_documents_to_queries(df1)
        self.assertEqual(set(['qid', 'query']), set(df2.columns))
        df1 = pt.new.ranked_documents([[1, 2], [2,0]], query=[["a a", "a a"], ["b b", "b b"]], stashed_results_0=[[1,1],[1,1,]])
        df2 = pt.model.ranked_documents_to_queries(df1)
        self.assertEqual(set(['qid', 'query', 'stashed_results_0']), set(df2.columns))        

    def test_doc_cols(self):
        df = pt.new.queries(["q1", "q2"])
        self.assertListEqual(['qid'], pt.model.document_columns(df))
        df2 = pt.new.ranked_documents([[1]], query=[["hello"]])
        self.assertEqual(set(['qid', 'docno', 'rank', 'score']), set(pt.model.document_columns(df2)))
        df2 = pt.new.ranked_documents([[1]], text=[["a"]], query=[["hello"]])
        self.assertEqual(set(['qid', 'docno', 'rank', 'score', "text"]), set(pt.model.document_columns(df2)))

    def test_query_cols(self):
        df = pt.new.queries(["q1", "q2"])
        self.assertEqual(2, len(df))
        df2 = pt.model.push_queries(df, keep_original=True)
        query_cols = pt.model.query_columns(df2)
        for col in ["qid", "query", "query_0"]:
            self.assertTrue(col in query_cols)

        df3 = pt.new.ranked_documents([[1]], query=[["hello"]])
        query_cols = pt.model.query_columns(df3)
        for col in ["qid", "query"]:
            self.assertTrue(col in query_cols)
        for col in ["docno", "score", "rank"]:
            self.assertFalse(col in query_cols)
        
        df = pd.DataFrame(columns=["qid", "query", "query_toks", "docno"])
        actual = pt.model.query_columns(df)
        expected = ["qid", "query", "query_toks"]
        self.assertEqual(actual, expected)

    def test_push_query(self):
        df = pt.new.queries(["q1", "q2"])
        self.assertEqual(2, len(df))

        df2 = pt.model.push_queries(df, keep_original=False)
        self.assertTrue("query_0" in df2.columns)
        self.assertFalse("query" in df2.columns)
        self.assertEqual("q1", df2.iloc[0]["query_0"])
        self.assertEqual("q2", df2.iloc[1]["query_0"])

        df_empty = df.head(0)
        df2_empty = pt.model.push_queries(df_empty, keep_original=False)
        self.assertTrue("query_0" in df2_empty.columns)
        self.assertFalse("query" in df2_empty.columns)

        df2 = pt.model.push_queries(df, keep_original=True)
        for col in ["query", "query_0"]:
            self.assertTrue(col in df2.columns)
            self.assertEqual("q1", df2.iloc[0][col])
            self.assertEqual("q2", df2.iloc[1][col])
        
        df3 = pt.model.push_queries(df2, keep_original=True)
        for col in ["query", "query_0", "query_1"]:
            self.assertTrue(col in df3.columns)
            self.assertEqual("q1", df3.iloc[0][col])
            self.assertEqual("q2", df3.iloc[1][col])
    
    def test_pop_query(self):
        df = pt.new.queries(["q1", "q2"])
        self.assertEqual(2, len(df))

        df2 = pt.model.push_queries(df, keep_original=False)
        self.assertTrue("query_0" in df2.columns)
        self.assertFalse("query" in df2.columns)
        df2["query"] = ["a", "b"]
        self.assertTrue("query" in df2.columns)
        self.assertEqual("q1", df2.iloc[0]["query_0"])
        self.assertEqual("q2", df2.iloc[1]["query_0"])

        df3 = pt.model.pop_queries(df2)
        self.assertFalse("query_1" in df3.columns)
        self.assertTrue("query" in df3.columns)
        # check that we dont have duplicated query column
        self.assertEqual(2, len(df3.columns))
        self.assertEqual("q1", df3.iloc[0]["query"])
        self.assertEqual("q2", df3.iloc[1]["query"])

    def test_rank_zero_query(self):
        df = pd.DataFrame([], columns=["qid", "docno", "score"])
        df = add_ranks(df)
        self.assertTrue("rank" in df.columns)

    def test_rank_one_query(self):
        for sq in [True, False]:
            df = pd.DataFrame([["q1", "doc1", 5], ["q1", "doc2", 5]], columns=["qid", "docno", "score"])
            df = add_ranks(df, single_query=True)
            self.assertTrue("rank" in df.columns)
            # check that first item is rank 1
            self.assertEqual(df.iloc[0]["rank"], FIRST_RANK)
            # check that ties are resolved by keeping the same order.
            # trec_eval instead breaks ties on ascending docno
            self.assertEqual(df.iloc[1]["rank"], FIRST_RANK+1)

    def test_rank_one_query_neg(self):
        for sq in [True, False]:
            df = pd.DataFrame([["q1", "doc1", -4], ["q1", "doc2", -5]], columns=["qid", "docno", "score"])
            df = add_ranks(df, single_query=sq)
            df = df.sort_values("rank", ascending=True)
            self.assertTrue("rank" in df.columns)
            # check that first item is rank 1
            self.assertEqual(df.iloc[0]["rank"], FIRST_RANK)
            self.assertEqual(df.iloc[0]["docno"], "doc1")

            df = pd.DataFrame([["q1", "doc2", -5], ["q1", "doc1", -4]], columns=["qid", "docno", "score"])
            df = add_ranks(df, single_query=sq)
            df = df.sort_values("rank", ascending=True)
            self.assertTrue("rank" in df.columns)
            # check that first item is rank 1
            self.assertEqual(df.iloc[0]["rank"], FIRST_RANK)
            self.assertEqual(df.iloc[0]["docno"], "doc1")
        
    def test_rank_one_query_sort(self):
        import pyterrier as pt
        sort_status = pt.model.STRICT_SORT
        pt.model.STRICT_SORT = True
        for sq in [True, False]:
            df = pd.DataFrame([["q1", "doc1", 4], ["q1", "doc2", 5]], columns=["qid", "docno", "score"])
            df = add_ranks(df, single_query=sq)
            self.assertTrue("rank" in df.columns)
            # check that first item is rank 1
            self.assertEqual(df.iloc[0]["rank"], FIRST_RANK)
            self.assertEqual(df.iloc[0]["docno"], "doc2")
        pt.model.STRICT_SORT = sort_status

    def test_rank_two_queries(self):
        df = pd.DataFrame([["q1", "doc1", 5], ["q1", "doc2", 4], ["q2", "doc1", 4]], columns=["qid", "docno", "score"])
        df = add_ranks(df)
        self.assertTrue("rank" in df.columns)
        self.assertEqual(df.iloc[0]["rank"], FIRST_RANK)
        self.assertEqual(df.iloc[1]["rank"], FIRST_RANK+1)
        self.assertEqual(df.iloc[2]["rank"], FIRST_RANK)

    def test_coerce_dataframe_with_string(self):
        input = "light"
        exp_result = pd.DataFrame([["1", "light"]], columns=['qid', 'query'])
        result = coerce_queries_dataframe(input)
        self.assertTrue(exp_result.equals(result))

    def test_coerce_dataframe_with_list(self):
        input = ["light", "mathematical", "electronic"]
        exp_result = pd.DataFrame([["1", "light"], ["2", "mathematical"], ["3", "electronic"]], columns=['qid', 'query'])
        result = coerce_queries_dataframe(input)
        self.assertTrue(exp_result.equals(result))

    def test_coerce_dataframe_throws_assertion_error(self):
        input = ("light", "mathematical", 25)
        self.assertRaises(AssertionError, coerce_queries_dataframe, input)

        input = {"25" : "mathematical"}
        self.assertRaises(ValueError, coerce_queries_dataframe, input)

        input = None
        self.assertRaises(ValueError, coerce_queries_dataframe, input)

    def test_coerce_dataframe_with_tuple(self):
        input = ("light", "mathematical", "electronic")
        exp_result = pd.DataFrame([["1", "light"], ["2", "mathematical"], ["3", "electronic"]], columns=['qid', 'query'])
        result = coerce_queries_dataframe(input)
        self.assertTrue(exp_result.equals(result))

    def test_coerce_dataframe_types(self):
        with self.subTest('typical'):
            input = pd.DataFrame([[1, 'query', 5, '1.3']], columns=['qid', 'query', 'docno', 'score'])
            exp_result = pd.DataFrame([['1', 'query', '5', 1.3]], columns=['qid', 'query', 'docno', 'score'])
            self.assertFalse(input.equals(exp_result))
            result = coerce_dataframe_types(input)
            pd.testing.assert_frame_equal(result, exp_result)
        with self.subTest('missing column'):
            input = pd.DataFrame([['query', 5, '1.3']], columns=['query', 'docno', 'score'])
            exp_result = pd.DataFrame([['query', '5', 1.3]], columns=['query', 'docno', 'score'])
            self.assertFalse(input.equals(exp_result))
            result = coerce_dataframe_types(input)
            pd.testing.assert_frame_equal(result, exp_result)
        with self.subTest('score not parsable as float'):
            input = pd.DataFrame([[1, 'query', 5, 'A']], columns=['qid', 'query', 'docno', 'score'])
            self.assertFalse(input.equals(exp_result))
            with self.assertRaises(ValueError):
                result = coerce_dataframe_types(input)

    def test_coerce_dataframe_types_torch(self):
        try:
            import torch
        except ImportError:
            self.skipTest("No torch installed")

        with self.subTest('score as integer'):
            input = pd.DataFrame([[1, 'query', 5, 1]], columns=['qid', 'query', 'docno', 'score'])
            exp_result = pd.DataFrame([['1', 'query', '5', 1.]], columns=['qid', 'query', 'docno', 'score'])
            self.assertFalse(input.equals(exp_result))
            result = coerce_dataframe_types(input)
            pd.testing.assert_frame_equal(result, exp_result)
        with self.subTest('score as torch type'):
            import torch
            input = pd.DataFrame([[1, 'query', 5, torch.tensor(1.3)]], columns=['qid', 'query', 'docno', 'score'])
            exp_result = pd.DataFrame([['1', 'query', '5', 1.3]], columns=['qid', 'query', 'docno', 'score'])
            self.assertFalse(input.equals(exp_result))
            result = coerce_dataframe_types(input)
            pd.testing.assert_frame_equal(result, exp_result)
        
    def test_split_Q(self):
        df = pt.new.queries(["a", "b", "c"])
        dfs = split_df(df, 2)
        self.assertEqual(2, len(dfs))
        self.assertEqual(2, len(dfs[0]))
        self.assertEqual(1, len(dfs[1]))

    def test_split_R(self):
        df = pt.new.ranked_documents([[2,1], [2]])
        dfs = split_df(df, 2)
        self.assertEqual(2, len(dfs))
        self.assertEqual(2, len(dfs[0]))
        self.assertEqual(1, len(dfs[1]))

        df = pt.new.ranked_documents([[2,1], [2,1], [2,1]])

        dfs = split_df(df, 2)
        self.assertEqual(2, len(dfs))
        self.assertEqual(4, len(dfs[0]))
        self.assertEqual(2, len(dfs[1]))
        



import pyterrier as pt

import itertools
import unittest
import tempfile
import shutil
import os
from unittest import SkipTest

from .base import TempDirTestCase, BaseTestCase

class TestIterDictIndexerPreTok(TempDirTestCase):

    def test_dpl(self):
        it1 = [
            {'docno': 'd1', 'url': 'url1', "toks" : {"a" : 1, "b" : 2}}
        ]
        it2 = [
            {'docno': 'd1', 'url': 'url1', "toks" : {"a" : 1, "b" : 2}},
            {'docno': 'd2', 'url': 'url1', "toks" : {"a" : 1, "b" : 1}}
        ]
        from pyterrier.terrier.index import DocListIterator
        
        iterator =  DocListIterator(iter(it1))
        self.assertTrue(iterator.hasNext())
        obj = iterator.next()
        self.assertIsNotNone(obj)
        self.assertEqual("d1", obj.getKey().get("docno"))
        self.assertEqual(2, obj.getValue().getFrequency("b"))
        self.assertEqual(3, obj.getValue().getDocumentLength())
        
        if iterator.hasNext():
            obj = iterator.next()
            self.assertIsNone(obj)
        self.assertFalse(iterator.hasNext())

        iterator =  DocListIterator(iter(it2))
        self.assertTrue(iterator.hasNext())
        obj = iterator.next()
        self.assertIsNotNone(obj)
        self.assertEqual("d1", obj.getKey().get("docno"))
        self.assertEqual(2, obj.getValue().getFrequency("b"))

        self.assertTrue(iterator.hasNext())
        obj = iterator.next()
        self.assertIsNotNone(obj)
        self.assertEqual(1, obj.getValue().getFrequency("b"))
        self.assertEqual("d2", obj.getKey().get("docno"))

        if iterator.hasNext():
            obj = iterator.next()
            self.assertIsNone(obj)
        self.assertFalse(iterator.hasNext())


    
    def test_json_pretok_iterator(self):
        if not pt.terrier.check_version("5.7") or not pt.terrier.check_helper_version("0.0.7"):
            raise SkipTest("Requires Terrier 5.7 and helper 0.0.7")

        it = [
            {'docno': 'd1', 'url': 'url1', "toks" : {"a" : 1, "b" : 2}}
        ]

        import json
        testfile = os.path.join(self.test_dir, "test.json")
        with open(testfile, 'wt') as file:
            for doc in it:
                file.write(json.dumps(doc))
        jparserCls = pt.java.autoclass("org.terrier.python.JsonlPretokenisedIterator")
        jparser = jparserCls(testfile)
        self.assertTrue(jparser.hasNext())
        nextRow = jparser.next()
        self.assertIsNotNone(nextRow)
        props = nextRow.getKey()
        self.assertEqual("d1", props.get("docno"))
        self.assertEqual("url1", props.get("url"))
        pl = nextRow.getValue()
        self.assertIn("a", pl.termSet())
        self.assertEqual(1, pl.getFrequency("a"))

    def _make_pretok_index(self, n, index_type, meta=('docno', 'url')):
        if not pt.terrier.check_version("5.7") or not pt.terrier.check_helper_version("0.0.7"):
            self.skipTest("Requires Terrier 5.7 and helper 0.0.7")
        
        from pyterrier.terrier.index import IndexingType
        # Test both versions: _fifo (for UNIX) and _nofifo (for Windows)
        indexers = [
            pt.index._IterDictIndexer_fifo(self.test_dir, type=index_type, meta=meta, pretokenised=True),
            pt.index._IterDictIndexer_fifo(self.test_dir, type=index_type, threads=4, meta=meta, pretokenised=True),
            pt.index._IterDictIndexer_nofifo(self.test_dir, type=index_type, meta=meta, pretokenised=True),
        ]
        if pt.utils.is_windows():
           indexers = [indexers[-1]] 
        for indexer in indexers:
            with self.subTest(indexer=indexer):
                it = [
                    {'docno': 'd1', 'url': 'url1', "toks" : {"a" : 1, "b" : 2.9123}},
                    {'docno': 'd2', 'url': 'url2', "toks" : {"a" : 1.5, "b" : 2.}}
                ]
                it = itertools.islice(it, n)
                print("Writing index to " + self.test_dir)
                indexref = indexer.index(it)
                self.assertIsNotNone(indexref)
                index = pt.IndexFactory.of(indexref)
                self.assertIsNotNone(index)
                self.assertEqual(n, index.getCollectionStatistics().getNumberOfDocuments())
                self.assertEqual(len(set(meta)), len(index.getMetaIndex().getKeys()))
                for m in meta:
                    self.assertTrue(m in index.getMetaIndex().getKeys())
                if index_type is IndexingType.CLASSIC:
                    self.assertTrue(os.path.isfile(self.test_dir + '/data.direct.bf'))
                    self.assertTrue(index.hasIndexStructure("direct"))
                else:
                    self.assertFalse(os.path.isfile(self.test_dir + '/data.direct.bf'))
                inv = index.getInvertedIndex()
                lex = index.getLexicon()
                post = inv.getPostings(lex.getLexiconEntry('a'))
                post.next()
                self.assertEqual(1, post.frequency)

                pindex = pt.java.cast("org.terrier.structures.PropertiesIndex", index)
                self.assertEqual("", pindex.getIndexProperty("termpipelines", "BLA"))

                index.close()
            # reset index directory for next run
            shutil.rmtree(self.test_dir)
            os.mkdir(self.test_dir)

    def test_pretok_createindex1_basic(self):
        from pyterrier.terrier.index import IndexingType
        from jnius import JavaException
        try:
            self._make_pretok_index(1, IndexingType.CLASSIC)
        except JavaException as ja:
            print("\n\t".join(ja.stacktrace))
            raise ja

    def test_pretok_createindex2_basic(self):
        from pyterrier.terrier.index import IndexingType
        self._make_pretok_index(2, IndexingType.CLASSIC)

    def test_pretok_createindex1_single_pass(self):
        from pyterrier.terrier.index import IndexingType
        from jnius import JavaException
        try:
            self._make_pretok_index(1, IndexingType.SINGLEPASS)
        except JavaException as ja:
            print("\n\t".join(ja.stacktrace))
            raise ja

    def test_pretok_createindex2_single_pass(self):
        from pyterrier.terrier.index import IndexingType
        from jnius import JavaException
        try:
            self._make_pretok_index(2, IndexingType.SINGLEPASS)
        except JavaException as ja:
            print("\n\t".join(ja.stacktrace))
            raise ja

if __name__ == "__main__":
    unittest.main()



import numpy as np
import pandas as pd
import pyterrier as pt
import os
import unittest
from .base import BaseTestCase
import tempfile
import shutil
import os

class TestApply(BaseTestCase):

    def test_drop_columns(self):
        from pyterrier.transformer import Transformer
        testDF = pd.DataFrame([["q1", "the bear and the wolf", 1]], columns=["qid", "query", "Bla"])
        p = pt.apply.Bla(drop=True)
        self.assertTrue(isinstance(p, Transformer))
        rtr = p(testDF)
        self.assertTrue("Bla" not in rtr.columns)

        rtr = p(testDF.to_dict(orient='records'))
        self.assertTrue("Bla" not in rtr[0])

    def test_index_apply(self):
        def _counter(iter_dict):
            count = 0
            for d in iter_dict:
                count += 1
            return count
        indexer = pt.apply.indexer(_counter)
        rtr = indexer.index([ {'docno' : 'd1'}, {'docno' : 'd2'}])
        self.assertEqual(2, rtr)

    def test_make_columns(self):
        from pyterrier.transformer import Transformer
        testDF = pd.DataFrame([["q1", "the bear and the wolf", 1]], columns=["qid", "query", "Bla"])
        p = pt.apply.BlaB(lambda row: row["Bla"] * 2)
        self.assertTrue(isinstance(p, Transformer))
        rtr = p(testDF)
        self.assertTrue("BlaB" in rtr.columns)
        self.assertEqual(rtr.iloc[0]["BlaB"], 2)
        emptyQs = pt.new.empty_Q()
        rtr = p(emptyQs)
        self.assertTrue("BlaB" in rtr.columns)

    def test_rename_columns(self):
        from pyterrier.transformer import Transformer
        testDF = pd.DataFrame([["q1", "the bear and the wolf", 1]], columns=["qid", "query", "Bla"])
        p = pt.apply.rename({'Bla' : "Bla2"})
        self.assertTrue(isinstance(p, Transformer))
        for input in [testDF, testDF.to_dict(orient='records')]:
            rtr = p(input)
            rtr = pd.DataFrame(list(rtr)) if not isinstance(rtr, pd.DataFrame) else rtr  # recover DF for easier testing
            self.assertTrue("Bla2" in rtr.columns)
            self.assertFalse("Bla" in rtr.columns)

        # test inspection
        self.assertEqual([["Bla"]], list(pt.inspect.transformer_inputs(p)))
        self.assertEqual(["qid", "query", "Bla2"], pt.inspect.transformer_outputs(p, ["qid", "query", "Bla"]))
        
        testDF2 = pd.DataFrame([["q1", "the bear and the wolf", 1]], columns=["qid", "query", "Bla2"])
        for input in [testDF2, testDF2.to_dict(orient='records')]:
            with self.assertRaises(KeyError):
                rtr = p(input)
        p_ignore = pt.apply.rename({'Bla' : "Bla2"}, errors='ignore')
        rtr = p_ignore(testDF2)

    def test_query_apply(self):
        stops=set(["and", "the"])
        origquery="the bear and the wolf"
        p = pt.apply.query(
                lambda q : " ".join([t for t in q["query"].split(" ") if not t in stops ])
            )
        testDF = pd.DataFrame([["q1", origquery]], columns=["qid", "query"])
        rtr = p(testDF)
        self.assertEqual(rtr.iloc[0]["query"], "bear wolf")
        self.assertEqual(rtr.iloc[0]["query_0"], origquery)

        testDF2 = pd.DataFrame([["q1"]], columns=["qid"])
        rtrDR2 = pt.apply.query(lambda row : row["qid"], required_columns=['qid'])(testDF2)
        self.assertEqual(rtrDR2.iloc[0]["query"], "q1")

    def test_query_apply_error(self):
        origquery="the bear and the wolf"
        testDF = pd.DataFrame([["q1", origquery]], columns=["qid", "query"])
        p = pt.apply.query(lambda q : q) # should thrown an error, as pt.apply.query should return a string, not a row
        with self.assertRaises(TypeError) as te:
            p(testDF)
        self.assertTrue("Could not coerce return from pt.apply.query function into a list of strings" in str(te.exception))

    def test_by_query_apply(self):
        inputDf = pt.new.ranked_documents([[1], [2]], qid=["1", "2"])
        def _inc_score(res):
            if len(res) == 0:
                return res
            res = res.copy()
            res["score"] = res["score"] + int(res.iloc[0]["qid"])
            return res
        p = pt.apply.by_query(_inc_score)
        outputDf = p(inputDf)
        self.assertEqual(outputDf.iloc[0]["qid"], "1")
        self.assertEqual(outputDf.iloc[0]["score"], 2)
        self.assertEqual(outputDf.iloc[1]["qid"], "2")
        self.assertEqual(outputDf.iloc[1]["score"], 4)

        outputDfEmpty = p(inputDf.head(0))

        p2 = pt.apply.by_query(lambda x: x)
        with self.assertRaisesRegex(ValueError, 'score column not present'):
            p2(pt.new.queries(['query 1', 'query 2']))

        p3 = pt.apply.by_query(lambda x: 5/0, add_ranks=False)
        with self.assertRaisesRegex(Exception, 'for qid 1'):
            p3(pt.new.queries(['query 1', 'query 2']))

    def test_by_query_apply_iter(self):
        inputDf = pt.new.ranked_documents([[1], [2]], qid=["1", "2"])
        def _inc_score(res):
            for row in res:
                row["score"] = row["score"] + int(row["qid"])
                yield row

        def _inc_score_list(res):
            rtr = []
            for row in res:
                row["score"] = row["score"] + int(row["qid"])
                rtr.append(row)
            return rtr
    
        for name, t in [
            ("gen", pt.apply.by_query(_inc_score, iter=True)),
            ("gen batch", pt.apply.by_query(_inc_score, iter=True, batch_size=1)),
            ("list", pt.apply.by_query(_inc_score_list, iter=True)),
            ("list batch", pt.apply.by_query(_inc_score_list, iter=True, batch_size=1))]:
            with self.subTest(name):
            
                outputDf = t(inputDf)
                self.assertEqual(outputDf.iloc[0]["qid"], "1")
                self.assertEqual(outputDf.iloc[0]["score"], 2)
                self.assertEqual(outputDf.iloc[1]["qid"], "2")
                self.assertEqual(outputDf.iloc[1]["score"], 4)

                # check no error on empty input
                outputDfEmpty = t(inputDf.head(0))

    def test_by_query_apply_batch(self):
        # same as test_by_query_apply, but batch_size is set.
        inputDf = pt.new.ranked_documents([[1], [2]], qid=["1", "2"])
        def _inc_score(res):
            if len(res) == 0:
                return res
            res = res.copy()
            res["score"] = res["score"] + int(res.iloc[0]["qid"])
            return res
        p = pt.apply.by_query(_inc_score, batch_size=1)
        outputDf = p(inputDf)
        self.assertEqual(outputDf.iloc[0]["qid"], "1")
        self.assertEqual(outputDf.iloc[0]["score"], 2)
        self.assertEqual(outputDf.iloc[1]["qid"], "2")
        self.assertEqual(outputDf.iloc[1]["score"], 4)

        outputDfEmpty = p(inputDf.head(0))

    def test_generic(self):
        inputDf = pt.new.ranked_documents([[1], [2]], qid=["1", "2"])
        def _fn1(df):
            df = df.copy()
            df["score"] = df["score"] * 2
            return df
        for i, t in enumerate([
            pt.apply.generic(_fn1),
            pt.apply.generic(_fn1, batch_size=1)
        ]):
            outputDf = t(inputDf)
            self.assertEqual(2, len(outputDf))
            self.assertEqual(outputDf.iloc[0]["qid"], "1")
            self.assertEqual(outputDf.iloc[0]["score"], 2)
            self.assertEqual(outputDf.iloc[1]["qid"], "2")
            self.assertEqual(outputDf.iloc[1]["score"], 4)
        
        def _fn2(df):
            df = df.copy()
            df["score"] = len(df)
            return df
        t1 = pt.apply.generic(_fn2)
        t2 = pt.apply.generic(_fn2, batch_size=1)
        outputDf1 = t1(inputDf)
        outputDf2 = t2(inputDf)
        self.assertEqual(2, len(outputDf1))
        self.assertEqual(2, len(outputDf2))
        # batch is the entire dataframe, ie 2 rows
        self.assertEqual(2, outputDf1.iloc[0]["score"])
        # batch is a one row dataframe
        self.assertEqual(1, outputDf2.iloc[0]["score"])
    
    def test_generic_iter(self):
        def _fn1_list(iter):
            rtr = []
            for i in iter:
                i["score"] *= 2
                rtr.append(i)
            return rtr
        
        def _fn1_gen(iter):
           for i in iter:
                i["score"] *= 2
                yield i

        for name, t in [
            ("list, no batch", pt.apply.generic(_fn1_list, iter=True)),
            ("list, batch", pt.apply.generic(_fn1_list, batch_size=1, iter=True)),
            ("gen, no batch", pt.apply.generic(_fn1_gen, iter=True)),
            ("gen, batch", pt.apply.generic(_fn1_gen, batch_size=1, iter=True)),
        ]:
            inputDf = pt.new.ranked_documents([[1], [2]], qid=["1", "2"])
            with self.subTest(name):
                outputDf1 = t(inputDf)
                self.assertEqual(2, len(outputDf1))
                self.assertEqual(2, outputDf1.iloc[0]["score"])
                self.assertEqual(4, outputDf1.iloc[1]["score"])

                outputDf1 = pd.DataFrame(t(inputDf.to_dict(orient='records')))
                self.assertEqual(2, len(outputDf1))
                self.assertEqual(2, outputDf1.iloc[0]["score"])
                self.assertEqual(4, outputDf1.iloc[1]["score"])

    
    def test_docscore_apply(self):
        p = pt.apply.doc_score(lambda doc_row: len(doc_row["text"]))
        testDF = pd.DataFrame([["q1", "hello", "d1", "aa"]], columns=["qid", "query", "docno", "text"])
        for input in [testDF, testDF.to_dict(orient='records')]:
            rtr = p(testDF)
            rtr = pd.DataFrame(list(rtr)) if not isinstance(rtr, pd.DataFrame) else rtr  # recover DF for easier testing
            self.assertEqual(rtr.iloc[0]["score"], 2.0)
            self.assertEqual(rtr.iloc[0]["rank"], pt.model.FIRST_RANK)
        
        rtr2 = p(testDF.head(0))
        self.assertTrue("rank" in rtr2.columns)
        self.assertTrue("score" in rtr2.columns)
        self.assertEqual('float64', rtr2['score'].dtype)

        rtr2 = p([])

    def test_docscore_batch(self):
        p = pt.apply.doc_score(lambda df: df["text"].str.len(), batch_size=2)
        testDF = pd.DataFrame([["q1", "hello", "d1", "aa"]], columns=["qid", "query", "docno", "text"])
        rtr = p(testDF)
        self.assertEqual(rtr.iloc[0]["score"], 2.0)
        self.assertEqual(rtr.iloc[0]["rank"], pt.model.FIRST_RANK)

        rtr2 = p(testDF.head(0))
        self.assertTrue("rank" in rtr2.columns)
        self.assertTrue("score" in rtr2.columns)
        self.assertEqual('float64', rtr2['score'].dtype)

    def test_docfeatures_apply(self):
        import numpy as np
        p = pt.apply.doc_features(lambda doc_row: np.array([0,1]) )
        testDF = pd.DataFrame([["q1", "hello", "d1", "aa"]], columns=["qid", "query", "docno", "text"])
        for input in [testDF, testDF.to_dict(orient='records')]:
            rtr = p(input)
            rtr = pd.DataFrame(list(rtr)) if not isinstance(rtr, pd.DataFrame) else rtr # recover DF for easier testing
            self.assertTrue(np.array_equal(rtr.iloc[0]["features"], np.array([0,1])))



@pt.testing.transformer_test_class
def test_apply_doc_features():
    return pt.apply.doc_features(lambda row: np.array([0,1]))


@pt.testing.transformer_test_class
def test_apply_doc_score():
    return pt.apply.doc_score(lambda row: 5)



import pandas as pd # an error occurring here ususally means failed pip installation not detected by GHA
import pyterrier as pt
import os
import unittest
import warnings
from .base import TempDirTestCase
import tempfile
import shutil

class TestFlash(TempDirTestCase):

    def test_one_row_round(self):
        import pyterrier as pt
        vaswani = pt.datasets.get_dataset("vaswani")
        br = pt.terrier.Retriever(vaswani.get_index())
        rtr = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg", "num_q", pt.measures.NDCG@5], round=2)
        self.assertEqual(str(rtr.iloc[0]["map"]), "0.31")
        self.assertEqual(str(rtr.iloc[0]["nDCG@5"]), "0.46")

        rtr = pt.Experiment([br], vaswani.get_topics().head(10), vaswani.get_qrels(), ["map", "ndcg", "num_q", pt.measures.NDCG@5], round={"nDCG@5" : 1})
        self.assertEqual(str(rtr.iloc[0]["nDCG@5"]), "0.5")


    def test_TREC_indexing(self):
        print("Writing index to " + self.test_dir)
        indexer = pt.TRECCollectionIndexer(self.test_dir)
        indexRef = indexer.index(pt.io.find_files(self.here + "/fixtures/vaswani_npl/corpus/"))
        self.assertIsNotNone(indexRef)
        index = pt.IndexFactory.of(indexRef)
        self.assertEqual(11429, index.getCollectionStatistics().getNumberOfDocuments())
        self.assertTrue(os.path.isfile(self.test_dir + '/data.direct.bf'))



import pandas as pd
from ir_measures import Measure
from typing import Literal, Union, Sequence, Callable, Tuple, IO, Dict, Any
from .. import Transformer
import types

MEASURE_TYPE=Union[str,Measure]
MEASURES_TYPE=Sequence[MEASURE_TYPE]
SAVEMODE_TYPE=Literal['reuse', 'overwrite', 'error', 'warn']
VALIDATE_TYPE = Literal['warn', 'error', 'ignore']

SYSTEM_OR_RESULTS_TYPE = Union[Transformer, pd.DataFrame]
SAVEFORMAT_TYPE = Union[Literal['trec'], types.ModuleType, Tuple[Callable[[IO], pd.DataFrame], Callable[[pd.DataFrame, IO], None]]]


NUMERIC_TYPE = Union[float,int,complex]
TEST_FN_TYPE = Callable[ [Sequence[NUMERIC_TYPE],Sequence[NUMERIC_TYPE]], Tuple[Any,NUMERIC_TYPE] ]

# we need types before imports relying on those types
from ._experiment import Experiment #noqa: E402
from ._grid import GridScan, GridSearch, KFoldGridSearch #noqa: E402

def Evaluate(res : pd.DataFrame, qrels : pd.DataFrame, metrics : MEASURES_TYPE= ['map', 'ndcg'], perquery : bool = False) -> Dict:
    """
    Evaluate a single result dataframe with the given qrels. This method may be used as an alternative to
    ``pt.Experiment()`` for getting only the evaluation measurements given a single set of existing results.

    The PyTerrier-way is to use ``pt.Experiment()`` to evaluate a set of transformers, but this method is useful
    if you have a set of results already, and want to evaluate them without having to create a transformer pipeline.

    :param res: Either a dataframe with columns=['qid', 'docno', 'score'] or a dict {qid:{docno:score,},}
    :param qrels: Either a dataframe with columns=['qid','docno', 'label'] or a dict {qid:{docno:label,},}
    :param metrics: A list of strings specifying which evaluation metrics to use. Default=['map', 'ndcg']
    :param perquery: If true return each metric for each query, else return mean metrics. Default=False
    """
    if len(res) == 0:
        raise ValueError("No results for evaluation")
    
    from ._execution import _run_and_evaluate
    _, rtr = _run_and_evaluate(res, None, qrels, metrics, perquery=perquery)
    return rtr

__all__ = ["Experiment", "Evaluate", "GridScan", "GridSearch", "KFoldGridSearch"]


"""Validation utilities for checking the inputs of transformers."""
import warnings
from types import TracebackType
from typing import List, Optional, Type, Union, Set
import pandas as pd
import pyterrier as pt


class _TransformerMode:
    def __init__(self, missing_columns: List[str], extra_columns: List[str], mode_name: Optional[str] = None):
        self.missing_columns = missing_columns
        self.extra_columns = extra_columns
        self.mode_name = mode_name

    def __str__(self):
        return f'{self.mode_name} (missing: {self.missing_columns}, extra: {self.extra_columns})'

    def __repr__(self):
        return f'TransformerMode(missing_columns={self.missing_columns!r}, ' \
               f'extra_columns={self.extra_columns!r}, ' \
               f'mode_name={self.mode_name!r})'


class InputValidationError(KeyError):
    """Exception raised when input validation of a single transformer fails.
    
    This is raised when the input DataFrame does not match the expected columns for a transformer.
    its typically raised by pt.validate.* methods.
    """
    def __init__(self, message: str, modes: List[_TransformerMode]):
        """Create an InputValidationError."""
        assert len(modes) > 0
        super().__init__(message)
        self.modes = modes

    def __str__(self):
        return self.args[0] + ' ' + str(self.modes)

    def __repr__(self):
        return f'InputValidationError({self.args[0]!r}, {self.modes!r})'


class InputValidationWarning(Warning):
    """Warning raised when input validation fails in warn mode.
    
    This is raised when the input DataFrame does not match the expected columns for a transformer,
    but the validation is in warn mode. It is typically produced by pt.validate.* methods."""
    pass


def columns(
    inp: Union[pd.DataFrame, List[str]],
    *,
    includes: Optional[List[str]] = None,
    excludes: Optional[List[str]] = None,
    warn: bool = False
) -> None:
    """Check that the input frame has the expected columns.

    Args:
        inp: Input DataFrame or columns to validate
        includes: List of required columns
        excludes: List of forbidden columns
        warn: If True, raise warnings instead of exceptions for validation errors

    Raises:
        InputValidationError: If warn=False and validation fails
        InputValidationWarning: If warn=True and validation fails

    .. versionchanged:: 0.15.0
        Accept ``List[str]`` inp columns
    """
    with any(inp, warn=warn) as v:
        v.columns(includes=includes, excludes=excludes)


def query_frame(inp: Union[pd.DataFrame, List[str]], extra_columns: Optional[List[str]] = None, warn: bool = False) -> None:
    """Check that the input frame is a valid query frame.

    Args:
        inp: Input DataFrame or columns to validate
        extra_columns: Additional required columns
        warn: If True, raise warnings instead of exceptions for validation errors

    Raises:
        InputValidationError: If warn=False and validation fails
        InputValidationWarning: If warn=True and validation fails

    .. versionchanged:: 0.15.0
        Accept ``List[str]`` inp columns
    """
    with any(inp, warn=warn) as v:
        v.query_frame(extra_columns)


def result_frame(inp: Union[pd.DataFrame, List[str]], extra_columns: Optional[List[str]] = None, warn: bool = False) -> None:
    """Check that the input frame is a valid result frame.

    Args:
        inp: Input DataFrame or columns to validate
        extra_columns: Additional required columns
        warn: If True, raise warnings instead of exceptions for validation errors

    Raises:
        InputValidationError: If warn=False and validation fails
        InputValidationWarning: If warn=True and validation fails

    .. versionchanged:: 0.15.0
        Accept ``List[str]`` inp columns
    """
    with any(inp, warn=warn) as v:
        v.result_frame(extra_columns)


def document_frame(inp: Union[pd.DataFrame, List[str]], extra_columns: Optional[List[str]] = None, warn: bool = False) -> None:
    """Check that the input frame is a valid document frame.

    Args:
        inp: Input DataFrame or columns to validate
        extra_columns: Additional required columns
        warn: If True, raise warnings instead of exceptions for validation errors

    Raises:
        InputValidationError: If warn=False and validation fails
        InputValidationWarning: If warn=True and validation fails

    .. versionchanged:: 0.15.0
        Accept ``List[str]`` inp columns
    """
    with any(inp, warn=warn) as v:
        v.document_frame(extra_columns)


def columns_iter(
    inp: 'pt.utils.PeekableIter',
    *,
    includes: Optional[List[str]] = None,
    excludes: Optional[List[str]] = None,
    warn: bool = False
) -> None:
    """Check that the input frame has the expected columns.

    Args:
        inp: Input DataFrame to validate
        includes: List of required columns
        excludes: List of forbidden columns
        warn: If True, raise warnings instead of exceptions for validation errors

    Raises:
        InputValidationError: If warn=False and validation fails
        InputValidationWarning: If warn=True and validation fails
    """
    with any_iter(inp, warn=warn) as v:
        v.columns(includes=includes, excludes=excludes)


def query_iter(inp: 'pt.utils.PeekableIter', extra_columns: Optional[List[str]] = None, warn: bool = False) -> None:
    """Check that the input iterator is a valid query iterator.

    Args:
        inp: Input iterator to validate
        extra_columns: Additional required columns
        warn: If True, raise warnings instead of exceptions for validation errors
    Raises:
        InputValidationError: If warn=False and validation fails
        InputValidationWarning: If warn=True and validation fails
    """
    with any_iter(inp, warn=warn) as v:
        v.query_iter(extra_columns)


def result_iter(inp: 'pt.utils.PeekableIter', extra_columns: Optional[List[str]] = None, warn: bool = False) -> None:
    """Check that the input iterator is a valid result iterator.

    Args:
        inp: Input iterator to validate
        extra_columns: Additional required columns
        warn: If True, raise warnings instead of exceptions for validation errors
    Raises:
        InputValidationError: If warn=False and validation fails
        InputValidationWarning: If warn=True and validation fails
    """
    with any_iter(inp, warn=warn) as v:
        v.result_iter(extra_columns)


def document_iter(inp: 'pt.utils.PeekableIter', extra_columns: Optional[List[str]] = None, warn: bool = False) -> None:
    """Check that the input iterator is a valid document iterator.

    Args:
        inp: Input iterator to validate
        extra_columns: Additional required columns
        warn: If True, raise warnings instead of exceptions for validation errors
    Raises:
        InputValidationError: If warn=False and validation fails
        InputValidationWarning: If warn=True and validation fails
    """
    with any_iter(inp, warn=warn) as v:
        v.document_iter(extra_columns)


def any(inp: Union[pd.DataFrame, List[str]], warn: bool = False) -> '_ValidationContextManager':
    """Create a validation context manager for a DataFrame or list of columns to test multiple possible modes.

    Args:
        inp: Input DataFrame or list of columns to validate
        warn: If True, raise warnings instead of exceptions for validation errors
    """
    return _ValidationContextManager(inp, warn=warn)


def any_iter(inp: 'pt.utils.PeekableIter', warn: bool = False) -> '_IterValidationContextManager':
    """Create a validation context manager for an iterator to test multiple possible modes.

    Args:
        inp: Input iterator to validate
        warn: If True, raise warnings instead of exceptions for validation errors
    """
    if not isinstance(inp, pt.utils.PeekableIter):
        raise AttributeError('inp is not peekable. Run the following before calling this function.\n'
                             'inp = pt.utils.peekable(inp) # !! IMPORTANT: you must re-assign the input to peekable '
                             '(not just pass it in), otherwise you will skip the first record !!')
    return _IterValidationContextManager(inp, warn=warn)


class _ValidationContextManager:
    """Context manager for validating the input to transformers."""
    def __init__(self, inp: Union[pd.DataFrame, List[str]], warn: bool = False):
        """Create a ValidationContextManager for the given DataFrame."""
        if isinstance(inp, pd.DataFrame):
            self.inp_columns = list(inp.columns)
        else:
            self.inp_columns = inp
        self.mode: Optional[str] = None
        self.attempts = 0
        self.errors: List[_TransformerMode] = []
        self.warn = warn

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType]
    ) -> Optional[bool]:

        if exc_type is not None:
            return False # the captured exception takes priority

        if self.attempts > 0 and self.attempts == len(self.errors):
            message = "DataFrame(columns=%s) does not match required columns for this transformer." % str(self.inp_columns)
            if self.warn:
                warnings.warn(f"{message} {self.errors}", InputValidationWarning)
            else:
                raise InputValidationError(message, self.errors)
        return None

    def columns(self,
                *,
                includes: Optional[List[str]] = None,
                excludes: Optional[List[str]] = None,
                mode: Optional[str] = None) -> bool:
        """Check that the input frame has the ``includes`` columns and doesn't have the ``excludes`` columns."""
        includes = includes if includes is not None else []
        excludes = excludes if excludes is not None else []
        missing_columns = set(includes) - set(self.inp_columns)
        extra_columns = set(excludes) & set(self.inp_columns)
        self.attempts += 1

        if missing_columns or extra_columns:
            self.errors.append(_TransformerMode(
                missing_columns=[c for c in includes if c in missing_columns],
                extra_columns=[c for c in excludes if c in extra_columns],
                mode_name=mode,
            ))
            return False

        if self.mode is None and mode is not None:
            self.mode = mode

        return True

    def query_frame(self, extra_columns: Optional[List[str]] = None, mode: Optional[str] = None) -> bool:
        """Check that the input frame is a valid query frame, with optional extra columns."""
        extra_columns = list(extra_columns) if extra_columns is not None else []
        return self.columns(includes=['qid'] + extra_columns, excludes=['docno'], mode=mode)

    def result_frame(self, extra_columns: Optional[List[str]] = None, mode: Optional[str] = None) -> bool:
        """Check that the input frame is a valid result frame, with optional extra columns."""
        extra_columns = list(extra_columns) if extra_columns is not None else []
        return self.columns(includes=['qid', 'docno'] + extra_columns, mode=mode)

    def document_frame(self, extra_columns: Optional[List[str]] = None, mode: Optional[str] = None) -> bool:
        """Check that the input frame is a valid document frame, with optional extra columns."""
        extra_columns = list(extra_columns) if extra_columns is not None else []
        return self.columns(includes=['docno'] + extra_columns, excludes=['qid'], mode=mode)


_EMPTY_ITER = object()

class _IterValidationContextManager:
    def __init__(self, inp: 'pt.utils.PeekableIter', warn: bool = False):
        self.sample_cols: Union[Set[str], object]
        try:
            self.sample_cols = set(inp.peek().keys())
        except StopIteration:
            self.sample_cols = _EMPTY_ITER
        self.mode: Optional[str] = None
        self.attempts = 0
        self.errors: List[_TransformerMode] = []
        self.warn = warn

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType]
    ) -> Optional[bool]:

        if exc_type is not None:
            return False # the captured exception takes priority

        if self.attempts > 0 and self.attempts == len(self.errors):
            message = "Input does not match required columns for this transformer."
            if self.warn:
                warnings.warn(f"{message} {self.errors}", InputValidationWarning)
            else:
                raise InputValidationError(message, self.errors)
        return None

    def columns(self,
                *,
                includes: Optional[List[str]] = None,
                excludes: Optional[List[str]] = None,
                mode: Optional[str] = None) -> bool:
        self.attempts += 1
        includes = includes if includes is not None else []
        excludes = excludes if excludes is not None else []
        if not isinstance(self.sample_cols, set):
            assert self.sample_cols == _EMPTY_ITER # AKA, but need to check type to make mypy happy
            self.errors.append(_TransformerMode(
                missing_columns=list(includes),
                extra_columns=[],
                mode_name=mode,
            ))
            return False
        missing_columns = set(includes) - self.sample_cols
        extra_columns = set(excludes) & self.sample_cols

        if missing_columns or extra_columns:
            self.errors.append(_TransformerMode(
                missing_columns=[c for c in includes if c in missing_columns],
                extra_columns=[c for c in excludes if c in extra_columns],
                mode_name=mode,
            ))
            return False

        if self.mode is None and mode is not None:
            self.mode = mode

        return True

    def query_iter(self, extra_columns: Optional[List[str]] = None, mode: Optional[str] = None) -> bool:
        """Check that the input iterator is a valid query iterator, with optional extra columns."""
        extra_columns = list(extra_columns) if extra_columns is not None else []
        return self.columns(includes=['qid'] + extra_columns, excludes=['docno'], mode=mode)

    def result_iter(self, extra_columns: Optional[List[str]] = None, mode: Optional[str] = None) -> bool:
        """Check that the input iterator is a valid result iterator, with optional extra columns."""
        extra_columns = list(extra_columns) if extra_columns is not None else []
        return self.columns(includes=['qid', 'docno'] + extra_columns, mode=mode)

    def document_iter(self, extra_columns: Optional[List[str]] = None, mode: Optional[str] = None) -> bool:
        """Check that the input iterator is a valid document iterator, with optional extra columns."""
        extra_columns = list(extra_columns) if extra_columns is not None else []
        return self.columns(includes=['docno'] + extra_columns, excludes=['qid'], mode=mode)

    def empty(self, *, mode: Optional[str] = 'empty'):
        self.attempts += 1
        if self.sample_cols != _EMPTY_ITER:
            self.errors.append(_TransformerMode(
                missing_columns=[],
                extra_columns=[],
                mode_name=mode,
            ))
            return False

        if self.mode is None and mode is not None:
            self.mode = mode
        return True



import pyterrier as pt
import tempfile
import unittest
from .base import TempDirTestCase
import os
import pandas as pd
import shutil

class TestIndexPipelines(TempDirTestCase):

    def test_add_dup(self):
        def _first(df):
            df2 = df.copy()
            df2["docno"] = df2["docno"] + "bis" 
            return pd.concat([df, df2])

        slider = pt.apply.generic(_first)
        indexer = pt.IterDictIndexer(self.test_dir)
        pipeline = slider >> indexer
        dataset = pt.get_dataset("irds:vaswani")
        count = 200
        indexref = pipeline.index(dataset.get_corpus_iter(count=count))
        self.assertIsNotNone(indexref)
        index = pt.IndexFactory.of(indexref)
        self.assertEqual( index.getCollectionStatistics().getNumberOfDocuments(),  2 * count)

    def test_sliding_title_two(self):
        corpus = [{"docno" : "d1", "text" : "A B C", "title" : "this is a title"}]
        slider = pt.text.sliding("text", 2, 1, prepend_attr="title")
        indexer = pt.IterDictIndexer(self.test_dir)
        pipeline = slider >> indexer
        indexref = pipeline.index(corpus)
        self.assertIsNotNone(indexref)
        index = pt.IndexFactory.of(indexref)
        # we should get 2 passages in the resulting index
        self.assertEqual(2, index.getCollectionStatistics().getNumberOfDocuments())

    def test_sliding_title_one(self):
        corpus = [{"docno" : "d1", "text" : "A B", "title" : "this is a title"}]
        slider = pt.text.sliding("text", 2, 1, prepend_attr="title")
        indexer = pt.IterDictIndexer(self.test_dir)
        pipeline = slider >> indexer
        indexref = pipeline.index(corpus)
        self.assertIsNotNone(indexref)
        index = pt.IndexFactory.of(indexref)
        # we should get 1 passages in the resulting index
        self.assertEqual(1, index.getCollectionStatistics().getNumberOfDocuments())


    def test_sliding(self):
        slider = pt.text.sliding("text", 10, 10, prepend_attr=None)
        indexer = pt.IterDictIndexer(self.test_dir)
        pipeline = slider >> indexer
        dataset = pt.get_dataset("irds:vaswani")
        count = 200
        indexref = pipeline.index(dataset.get_corpus_iter(count=count))
        self.assertIsNotNone(indexref)
        index = pt.IndexFactory.of(indexref)
        self.assertTrue( index.getCollectionStatistics().getNumberOfDocuments() > count)



import pandas as pd
import unittest
import pyterrier as pt
import inspect
from .base import BaseTestCase
from functools import partial

class TestInspect(BaseTestCase):

    def assertSortedEquals(self, arr1, arr2):
        self.assertEqual(sorted(arr1), sorted(arr2))

    def assertInNoOrder(self, member, container):
        container_set = [set(item) for item in container]
        self.assertIn(set(member), container_set)     

    def test_inspection_mincols(self):
        op_input = [[['qid', 'docno']]]
        sub_inputs = [
            [['qid', 'docno', 'query']],
            [['qid', 'docno', 'query']]
        ]
        plausible_configs = pt.inspect._minimal_inputs(op_input + sub_inputs)
        self.assertInNoOrder(['qid', 'docno', 'query'], plausible_configs)


    def test_terrier_retriever(self):
        JIR = pt.java.autoclass('org.terrier.querying.IndexRef')
        indexref = JIR.of(self.here + "/fixtures/index/data.properties")
        retr = pt.terrier.Retriever(indexref)
        textl = pt.text.get_text(pt.get_dataset('irds:vaswani'), "text")

        inputs = pt.inspect.transformer_inputs(retr)
        self.assertEqual(['qid', 'query'], inputs[0])
        self.assertIn(['qid', 'query', 'docid'], inputs)
        self.assertIn(['qid', 'docno', 'query'], inputs)
        self.assertIn(['qid', 'query_toks'], inputs)

        # with a text loader:
        inputs = pt.inspect.transformer_inputs(retr >> textl)
        self.assertEqual(['qid', 'query'], inputs[0])
        self.assertIn(['qid', 'query', 'docid'], inputs)
        self.assertIn(['qid', 'docno', 'query'], inputs)
        self.assertIn(['qid', 'query_toks'], inputs)

        # as a retriever
        rank_cols = pt.inspect.transformer_outputs(retr, ["qid", "query"])
        example_res = retr.search("what are chemical reactions")
        self.assertSortedEquals(rank_cols, example_res.columns.tolist())

        # as a reranker
        rerank_cols = pt.inspect.transformer_outputs(retr, rank_cols)
        example_res = retr.transform(example_res)
        self.assertSortedEquals(rerank_cols, example_res.columns.tolist())

        # with prf
        rm3 = pt.terrier.rewrite.RM3(indexref)
        prf_cols = pt.inspect.transformer_outputs(rm3, rank_cols)
        example_res = rm3.transform(example_res)
        self.assertSortedEquals(prf_cols, example_res.columns.tolist())

        # as a feature pipeline
        feat_pipe = retr >> (retr ** retr)
        feat_cols = pt.inspect.transformer_outputs(feat_pipe, ["qid", "query"])
        example_res = feat_pipe.search("what are chemical reactions")
        self.assertSortedEquals(feat_cols, example_res.columns.tolist())

        # with a text loader:
        textl_pipe = retr >> textl
        text_cols = pt.inspect.transformer_outputs(textl_pipe, ["qid", "query"])
        self.assertSortedEquals(rank_cols + ["text"], text_cols)

        # in a linear combination
        linear_pipe = retr + retr
        linear_cols = pt.inspect.transformer_outputs(linear_pipe, ["qid", "query"])
        example_res = linear_pipe.search("what are chemical reactions")
        self.assertSortedEquals(linear_cols, example_res.columns.tolist())

        # in a set combination
        intersect_pipe = retr & retr
        intersect_cols = pt.inspect.transformer_outputs(intersect_pipe, ["qid", "query"])
        example_res = intersect_pipe.search("what are chemical reactions")
        self.assertSortedEquals(intersect_cols, example_res.columns.tolist())

        subtransformers = pt.inspect.subtransformers(retr)
        self.assertEqual({}, subtransformers)

        attributes = pt.inspect.transformer_attributes(retr)
        self.assertEqual([
            pt.inspect.TransformerAttribute('index_location', retr.indexref),
            pt.inspect.TransformerAttribute('num_results', 1000),
            pt.inspect.TransformerAttribute('metadata', retr.metadata),
            pt.inspect.TransformerAttribute('wmodel', retr.controls['wmodel']),
            pt.inspect.TransformerAttribute('threads', retr.threads),
            pt.inspect.TransformerAttribute('verbose', retr.verbose),
            pt.inspect.TransformerAttribute('terrierql', retr.controls['terrierql']),
            pt.inspect.TransformerAttribute('parsecontrols', retr.controls['parsecontrols']),
            pt.inspect.TransformerAttribute('parseql', retr.controls['parseql']),
            pt.inspect.TransformerAttribute('applypipeline', retr.controls['applypipeline']),
            pt.inspect.TransformerAttribute('localmatching', retr.controls['localmatching']),
            pt.inspect.TransformerAttribute('filters', retr.controls['filters']),
            pt.inspect.TransformerAttribute('decorate', retr.controls['decorate']),
            pt.inspect.TransformerAttribute('decorate_batch', retr.controls['decorate_batch']),
            pt.inspect.TransformerAttribute('querying.processes', retr.properties['querying.processes']),
            pt.inspect.TransformerAttribute('querying.postfilters', retr.properties['querying.postfilters']),
            pt.inspect.TransformerAttribute('querying.default.controls', retr.properties['querying.default.controls']),
            pt.inspect.TransformerAttribute('querying.allowed.controls', retr.properties['querying.allowed.controls']),
            pt.inspect.TransformerAttribute('termpipelines', retr.properties['termpipelines']),
        ], attributes)

        new_retr = pt.inspect.transformer_apply_attributes(retr, num_results=10)
        self.assertEqual(new_retr.controls['end'], '9')

        subtransformers = pt.inspect.subtransformers(retr >> textl)
        self.assertEqual({'transformers': [retr, textl]}, subtransformers)

        attributes = pt.inspect.transformer_attributes(retr >> textl)
        self.assertEqual(attributes, [
            pt.inspect.TransformerAttribute('transformers', (retr, textl), init_parameter_kind=inspect.Parameter.VAR_POSITIONAL)
        ], attributes)


    def test_ltr_pipeline(self):
        import numpy as np

        feat_res = pd.DataFrame([
            ["q1", "a", np.array([1,2])],
            ["q2", "a", np.array([2, 5])],
            ["q1", "b", np.array([3, 5])],
            ["q2", "b", np.array([4, 4])],
            ["q1", "c", np.array([5, 0])]
        ], columns=["qid", "docno", "features"])
        qrels = pd.DataFrame([
            ["q1", "a", 1],
            ["q1", "b", 0],
            ["q2", "a", 1],
            ["q2", "b", 0]
        ], columns=["qid", "docno", "label"])

        from sklearn.ensemble import RandomForestClassifier
        rf = RandomForestClassifier()
        

        pipelines = [
            pt.ltr.apply_learned_model(rf),
            pt.ltr.ablate_features(1) >> pt.ltr.apply_learned_model(rf),
            pt.ltr.keep_features(0) >> pt.ltr.apply_learned_model(rf)
        ]
        try:
            import xgboost as xgb
            xgparams = {
                'objective': 'rank:ndcg',
                'learning_rate': 0.1,
                'gamma': 1.0, 'min_child_weight': 0.1,
                'max_depth': 6,
                'random_state': 42
            }
            pipelines.append(pt.ltr.apply_learned_model(xgb.XGBRanker(**xgparams), form="ltr"))
        except Exception as e:
            print("xgboost not installed, skipping xgboost pipeline test", e)
            pass

        try:
            import fastrank
            train_request = fastrank.TrainRequest.coordinate_ascent()
            params = train_request.params
            params.init_random = True
            params.normalize = True
            params.seed = 1234567
            pipelines.append(pt.ltr.apply_learned_model(train_request, form="fastrank"))
        except Exception as e: 
            print("fastrank not installed, skipping xgboost pipeline test", e)
            pass

        print("testing %d pipelines" % len(pipelines))
        for pipeline in pipelines:
            with self.subTest(pipeline=pipeline):
                print("Testing pipeline:", pipeline)
                pipeline.fit(feat_res, qrels, feat_res, qrels)
                result_res = pipeline.transform(feat_res)

                ltr_cols = pt.inspect.transformer_outputs(pipeline, feat_res.columns.tolist())
                self.assertSortedEquals(ltr_cols, result_res.columns.tolist())

    def test_rewrite_query(self):
        df = pd.DataFrame({
            'qid': ['1', '2'],
            'query': ['query1', 'query2'],
        })
        t = pt.apply.query(lambda x: x["query"] + " context")
        cols = pt.inspect.transformer_outputs(t, df.columns.tolist())
        self.assertSortedEquals(cols, t(df).columns.tolist())

    def test_apply_docscore(self):
        df = pd.DataFrame({
            'qid': ['1', '2'],
            'query': ['query1', 'query2'],
            'docno': ['doc1', 'doc2']
        })
        t = pt.apply.doc_score(lambda x: 1)
        cols = pt.inspect.transformer_outputs(t, df.columns.tolist())
        self.assertIn('score', cols)
        self.assertIn('rank', cols)
        self.assertSortedEquals(cols, t(df).columns.tolist())

    def test_apply_docfeatures(self):
        import numpy as np
        df = pd.DataFrame({
            'qid': ['1', '2'],
            'query': ['query1', 'query2'],
            'docno': ['doc1', 'doc2']
        })
        t = pt.apply.doc_features(lambda x: np.array([1.0, 2.0]))
        cols = pt.inspect.transformer_outputs(t, df.columns.tolist())
        self.assertIn('features', cols)
        self.assertSortedEquals(cols, t(df).columns.tolist())

    def test_apply_mkcol(self):
        df = pd.DataFrame({
            'qid': ['1', '2'],
            'query': ['query1', 'query2'],
            'docno': ['doc1', 'doc2']
        })
        t = pt.apply.newcol(lambda x: x['query'] + " context")
        cols = pt.inspect.transformer_outputs(t, df.columns.tolist())
        self.assertIn('newcol', cols)
        self.assertSortedEquals(cols, t(df).columns.tolist())

    def test_apply_dropcol(self):
        df = pd.DataFrame({
            'qid': ['1', '2'],
            'query': ['query1', 'query2'],
            'docno': ['doc1', 'doc2']
        })
        t = pt.apply.docno(drop=True)
        cols = pt.inspect.transformer_outputs(t, df.columns.tolist())
        self.assertNotIn('docno', cols)
        self.assertSortedEquals(cols, t(df).columns.tolist())

    def test_apply_generic(self):
        self._apply_generic(pt.apply.generic)

    def test_apply_byquery(self):
        self._apply_generic(partial(pt.apply.by_query, add_ranks=False))

    def _apply_generic(self, applyfn):
        df = pd.DataFrame({
            'qid': ['1', '2'],
            'query': ['query1', 'query2'],
            'docno': ['doc1', 'doc2']
        })

        # Define a generic function that adds a new column based on existing columns
        # this assugnment works with an empty DataFrame
        def _generic_func(df):
            return df.assign(newcol=df['query'] + " context")

        t = applyfn(_generic_func)
        cols = pt.inspect.transformer_outputs(t, df.columns.tolist())
        self.assertIn('newcol', cols)
        self.assertSortedEquals(cols, t(df).columns.tolist())

        def _generic_func_noempty(df):
            if len(df) == 0:
                return df # bad behavior, but we want to empty dfs to have the same return columns
            return df.assign(newcol=df['query'] + " context")

        t = applyfn(_generic_func_noempty)
        cols = pt.inspect.transformer_outputs(t, df.columns.tolist())
        self.assertNotIn('newcol', cols)
        
        # undesired behavior: raises error on empty DataFrame - cannot be inspected
        def _generic_func_empty(df):
            if len(df) == 0:
                raise ValueError("Empty DataFrame")
            return df.assign(newcol=df['query'] + " context")
        
        t = applyfn(_generic_func_empty)
        with self.assertRaises(pt.inspect.InspectError):
            cols = pt.inspect.transformer_outputs(t, df.columns.tolist())

        # now check we can specify the columns when specifying transform_outputs
        t = applyfn(_generic_func_empty, transform_outputs=lambda cols: cols + ['newcol'])
        cols = pt.inspect.transformer_outputs(t, df.columns.tolist())
        self.assertIn('newcol', cols)
        self.assertSortedEquals(cols, t(df).columns.tolist())

        # same for the erroring function
        t = applyfn(_generic_func_noempty, transform_outputs=lambda cols: cols + ['newcol'])
        cols = pt.inspect.transformer_outputs(t, df.columns.tolist())
        self.assertIn('newcol', cols)
        self.assertSortedEquals(cols, t(df).columns.tolist())

    def test_apply_generic_iter(self):
        self._apply_generic_iter(pt.apply.generic)

    def test_apply_byquery_iter(self):
        self._apply_generic_iter(pt.apply.by_query)

    def _apply_generic_iter(self, applyfn):
        df = pd.DataFrame({
            'qid': ['1', '2'],
            'query': ['query1', 'query2'],
            'docno': ['doc1', 'doc2']
        })

        # Define a generic function that adds a new column based on existing columns
        def _generic_func_iter(iter):
            for r in iter:
                r["newcol"] = r['query'] + " context"
                yield r

        # an iter-only transformer is not inspectable
        t = applyfn(_generic_func_iter, iter=True)
        with self.assertRaises(pt.inspect.InspectError) as ie:
            cols = pt.inspect.transformer_outputs(t, df.columns.tolist())
        self.assertIn('not inspectable', str(ie.exception.__cause__))

        t = applyfn(_generic_func_iter, iter=True, transform_outputs=lambda cols: cols + ['newcol'])
        cols = pt.inspect.transformer_outputs(t, df.columns.tolist())
        self.assertIn('newcol', cols)
        self.assertSortedEquals(cols, t(df).columns.tolist())
    
    def test_rename(self):
        df = pd.DataFrame({
            'qid': ['1', '2'],
            'query': ['query1', 'query2'],
            'context': ['context1', 'context2']
        })
        br0 = pt.Transformer.from_df(df[['qid', 'query']], uniform=True)
        br1 = pt.Transformer.from_df(df, uniform=True)

        cols = pt.inspect.transformer_outputs(br1 >> pt.apply.rename({'context' : 'prompt'}), ["qid", "query", "context"])
        self.assertSortedEquals(cols, ['qid', 'query', 'prompt'])

        with self.assertRaises(pt.validate.InputValidationError):
            pt.inspect.transformer_outputs(br0 >> pt.apply.rename({'context' : 'prompt'}), ["qid", "query"])

    def test_rename_nocols(self):
        df = pd.DataFrame({
            'qid': ['1', '2'],
            'query': ['query1', 'query2'],
            'context': ['context1', 'context2']
        })
        br0 = pt.Transformer.from_df(df[['qid', 'query']], uniform=True)
        br1 = pt.Transformer.from_df(df, uniform=True)

        def _rename_context(df):
            if len(df) == 0:
                raise ValueError("Empty DataFrame")
            return df.rename(columns={'context' : 'prompt'})

        with self.assertRaises(pt.inspect.InspectError):
            pt.inspect.transformer_outputs(br1 >> pt.apply.generic(_rename_context), ["qid", "query", "context"])

        with self.assertRaises(pt.inspect.InspectError):
            pt.inspect.transformer_outputs(br0 >> pt.apply.generic(_rename_context), ["qid", "query"])

    def test_subtransformers_bad_attr_inspect(self):

        # this class does not set self.x in __init__  
        class A(pt.Transformer):
            def __init__(self, x):
                # x is ignored
                super().__init__()

            def transform(self, inp):
                return None
        
        with self.assertRaises(pt.inspect.InspectError):
            print(pt.inspect.transformer_attributes(A(1)))

        attrs = pt.inspect.transformer_attributes(A(1), strict=False)
        self.assertEqual(1, len(attrs))
        self.assertEqual('x', attrs[0].name)
        self.assertEqual(pt.inspect.TransformerAttribute.MISSING, attrs[0].value)

        self.assertEqual(0, len(pt.inspect.subtransformers(A(1))))

    def test_transformer_type(self):
        class A(pt.Transformer):
            def transform(self, inp):
                pass
        class B(pt.Indexer):
            def index(self, inp):
                pass
        class C(pt.Indexer):
            def transform(self, inp):
                pass
            def index(self, inp):
                pass

        self.assertEqual(pt.inspect.transformer_type(A()), pt.inspect.TransformerType.transformer)
        self.assertEqual(pt.inspect.transformer_type(B()), pt.inspect.TransformerType.indexer)
        self.assertEqual(pt.inspect.transformer_type(C()), pt.inspect.TransformerType.transformer | pt.inspect.TransformerType.indexer)
        self.assertEqual(pt.inspect.transformer_type(object()), pt.inspect.TransformerType(0))

if __name__ == "__main__":
    unittest.main()
        



"""DEPRECATED MODULE

Warning: This module is deprecated and will be removed in a future version.

It isn't loaded by __init__ -- it's intended to handle cases where users:
import pyterrier.index
or
from pyterrier.index import X
"""

from deprecated import deprecated
from enum import Enum
import pyterrier as pt

@deprecated(version='0.11.0', reason="use pt.terrier.IterDictIndexer() instead")
class IterDictIndexer(pt.terrier.IterDictIndexer): # type: ignore[valid-type,misc] # IterDictIndexer can be either fifo or non-fifo version
    pass

@deprecated(version='0.11.0', reason="use pt.terrier.DFIndexer() instead")
class DFIndexer(pt.terrier.DFIndexer):
    pass

@deprecated(version='0.11.0', reason="use pt.terrier.TRECCollectionIndexer() instead")
class TRECCollectionIndexer(pt.terrier.DFIndexer):
    pass

@deprecated(version='0.11.0', reason="use pt.terrier.FilesIndexer() instead")
class FilesIndexer(pt.terrier.DFIndexer):
    pass

# Enumerations are tricky since we can't subclass them :(. Just duplicate while deprecated

@deprecated(version='0.11.0', reason="use pt.terrier.IndexingType() instead")
class IndexingType(Enum):
    CLASSIC = 1
    SINGLEPASS = 2
    MEMORY = 3



from deprecated import deprecated
import pyterrier as pt

@deprecated(version='1.0.0', reason="use pyterrier_alpha.fusion.PerQueryMaxMinScore() instead")
class PerQueryMaxMinScoreTransformer(pt.Transformer):
    '''
    applies per-query maxmin scaling on the input scores
    '''
    
    def transform(self, topics_and_res):
        from sklearn.preprocessing import minmax_scale
        topics_and_res = topics_and_res.copy()
        topics_and_res["score"] = topics_and_res.groupby('qid')["score"].transform(lambda x: minmax_scale(x))
        return topics_and_res

Experiment = deprecated(version='1.0.0', reason="use pt.Experiment() instead")(pt.Experiment)
Evaluate = deprecated(version='1.0.0', reason="use pt.Evaluate() instead")(pt.Experiment)
GridScan = deprecated(version='1.0.0', reason="use pt.GridScan() instead")(pt.GridScan)
GridSearch = deprecated(version='1.0.0', reason="use pt.GridSearch() instead")(pt.GridScan)
KFoldGridSearch = deprecated(version='1.0.0', reason="use pt.KFoldGridSearch() instead")(pt.KFoldGridSearch)
